# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2014-2015 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu


class TestQConversion(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.mat = xu.materials.Si
        cls.hxrd = xu.HXRD(cls.mat.Q(1, 1, 0), cls.mat.Q(0, 0, 1))
        cls.nch = 9
        cls.ncch = 4
        cls.hxrd.Ang2Q.init_linear('z+', cls.ncch, cls.nch, 1.0, 50e-6, 0.)
        cls.hklsym = (0, 0, 4)
        cls.hklasym = (2, 2, 4)

    def test_qconversion_linear(self):
        ang = self.hxrd.Q2Ang(self.mat.Q(self.hklsym))
        qout = self.hxrd.Ang2HKL(ang[0], ang[3], mat=self.mat,
                                 dettype='linear')
        self.assertEqual(qout[0].size, self.nch)
        for i in range(3):
            q = qout[i]
            self.assertAlmostEqual(numpy.average(q), self.hklsym[i], places=6)

    def test_qconversion_linear_asym(self):
        ang = self.hxrd.Q2Ang(self.mat.Q(self.hklasym))
        qout = self.hxrd.Ang2HKL(ang[0], ang[3], mat=self.mat,
                                 dettype='linear')
        self.assertEqual(qout[0].size, self.nch)
        for i in range(3):
            q = qout[i]
            self.assertAlmostEqual(numpy.average(q), self.hklasym[i], places=6)

    def test_qconversion_linear_energy(self):
        ang1 = self.hxrd.Q2Ang(self.mat.Q(self.hklsym))
        ang2 = self.hxrd.Q2Ang(self.mat.Q(self.hklsym) / 2.)
        qout = self.hxrd.Ang2HKL((ang1[0], ang2[0]), (ang1[3], ang2[3]),
                                 en=(self.hxrd.energy, 2 * self.hxrd.energy),
                                 mat=self.mat, dettype='linear')
        for i in range(3):
            q = qout[i]
            self.assertAlmostEqual(q[0, self.ncch], self.hklsym[i], places=10)
            self.assertAlmostEqual(q[1, self.ncch], self.hklsym[i], places=10)

    def test_qconversion_linear_detpos(self):
        tt = numpy.random.rand() * 90
        dpos = self.hxrd.Ang2Q.getDetectorPos(tt, dim=1)
        ki = self.hxrd._A2QConversion.r_i / \
            numpy.linalg.norm(self.hxrd._A2QConversion.r_i) * self.hxrd.k0
        qout = self.hxrd.Ang2Q.linear(0, tt)
        for j, x, y, z in zip(range(self.nch), dpos[0], dpos[1], dpos[2]):
            vpos = numpy.asarray((x, y, z))
            kf = vpos / numpy.linalg.norm(vpos) * self.hxrd.k0
            for i in range(3):
                self.assertAlmostEqual(qout[i][j], kf[i]-ki[i], places=10)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2012 Dominik Kriegner <dominik.kriegner@gmail.com>

import numpy
import xrayutilities as xu

Si = xu.materials.Si  # load material from materials submodule

# initialize experimental class with directions from experiment
exp = xu.HXRD(Si.Q(1, 1, -2), Si.Q(1, 1, 1))

# calculate angles and print them to the screen
angs = exp.Q2Ang(Si.Q(1, 1, 1))
print("|F000|: %8.3f"
      % (numpy.abs(Si.StructureFactor(Si.Q(0, 0, 0), exp.energy))))

print("Si (111)")
print("phi:%8.4f" % angs[2])
print("om: %8.4f" % angs[0])
print("tt: %8.4f" % angs[3])
print("|F|: %8.3f"
      % (numpy.abs(Si.StructureFactor(Si.Q(1, 1, 1), exp.energy))))


angs = exp.Q2Ang(Si.Q(2, 2, 4))
print("Si (224)")
print("phi:%8.4f" % angs[2])
print("om: %8.4f" % angs[0])
print("tt: %8.4f" % angs[3])
print("|F|: %8.3f"
      % (numpy.abs(Si.StructureFactor(Si.Q(2, 2, 4), exp.energy))))


angs = exp.Q2Ang(Si.Q(3, 3, 1))
print("Si (331)")
print("phi:%8.4f" % angs[2])
print("om: %8.4f" % angs[0])
print("tt: %8.4f" % angs[3])
print("|F|: %8.3f"
      % (numpy.abs(Si.StructureFactor(Si.Q(3, 3, 1), exp.energy))))



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2019 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu

try:
    import lmfit
except ImportError:
    lmfit = None


@unittest.skipIf(lmfit is None, "the lmfit Python package is needed")
class Test_DynamicalModel(unittest.TestCase):
    # define used layer stack
    sub = xu.simpack.Layer(xu.materials.GaAs, numpy.inf)
    lay = xu.simpack.Layer(xu.materials.AlGaAs(0.75), 995.64, relaxation=0.0)
    pls = xu.simpack.PseudomorphicStack001('AlGaAs on GaAs', sub, lay)
    hkl = (0, 0, 4)

    # simulation parameters
    kwargs = dict(I0=1e6, background=0, resolution_width=1e-5)

    @classmethod
    def setUpClass(cls):
        cls.sdyn = xu.simpack.SimpleDynamicalCoplanarModel(cls.pls,
                                                           **cls.kwargs)
        cls.dyn = xu.simpack.DynamicalModel(cls.pls, **cls.kwargs)
        cls.fms = xu.simpack.FitModel(cls.sdyn)
        cls.fmd = xu.simpack.FitModel(cls.dyn)
        qz = numpy.linspace(4.40, 4.50, 2000)
        cls.ai = xu.simpack.coplanar_alphai(0, qz)

    def test_Calculation(self):
        sim = self.dyn.simulate(self.ai, hkl=self.hkl)
        self.assertEqual(len(sim), len(self.ai))
        self.assertTrue(numpy.all(sim >= self.kwargs['background']))
        self.assertTrue(numpy.all(sim <= self.kwargs['I0']))

    def test_FitModel_eval(self):
        sim1 = self.sdyn.simulate(self.ai, hkl=self.hkl)
        p = self.fms.make_params()
        sim2 = self.fms.eval(p, x=self.ai, hkl=self.hkl)
        for v1, v2 in zip(sim1, sim2):
            self.assertAlmostEqual(v1, v2, places=10)

    def test_FitModel_eval_dyn(self):
        sim1 = self.dyn.simulate(self.ai, hkl=self.hkl)
        p = self.fmd.make_params()
        sim2 = self.fmd.eval(p, x=self.ai, hkl=self.hkl)
        for v1, v2 in zip(sim1, sim2):
            self.assertAlmostEqual(v1, v2, places=10)

    def test_Consistency(self):
        sim1 = self.sdyn.simulate(self.ai, hkl=self.hkl)
        sim2 = self.dyn.simulate(self.ai, hkl=self.hkl)
        self.assertEqual(len(sim1), len(sim2))
        self.assertEqual(numpy.argmax(sim1), numpy.argmax(sim2))
        self.assertAlmostEqual(xu.math.fwhm_exp(self.ai, sim1),
                               xu.math.fwhm_exp(self.ai, sim2), places=4)
        self.assertTrue(xu.math.fwhm_exp(self.ai, sim1) >=
                        self.kwargs['resolution_width'])


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2014 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import xrayutilities as xu


class TestQ2Ang_nonCOP(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.mat = xu.materials.Al2O3
        cls.hxrd = xu.NonCOP(cls.mat.Q(1, 1, 0), cls.mat.Q(0, 0, 1))
        cls.hkltest = (1, 2, 3)
        cls.hkltest2 = (-1.5, 3.1, 3)

    def test_Q2Ang_nonCOP_point(self):
        ang = self.hxrd.Q2Ang(self.mat.Q(self.hkltest))
        qout = self.hxrd.Ang2HKL(*ang, mat=self.mat)
        for i in range(3):
            self.assertAlmostEqual(qout[i], self.hkltest[i], places=10)

    def test_Q2Ang_nonCOP_point2(self):
        ang = self.hxrd.Q2Ang(self.mat.Q(self.hkltest2))
        qout = self.hxrd.Ang2HKL(*ang, mat=self.mat)
        for i in range(3):
            self.assertAlmostEqual(qout[i], self.hkltest2[i], places=10)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009-2010 Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2009-2019 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
a class for observing a SPEC data file

Motivation:

SPEC files can become quite large. Therefore, subsequently reading
the entire file to extract a single scan is a quite cumbersome procedure.
This module is a proof of concept code to write a file observer starting
a reread of the file starting from a stored offset (last known scan position)
"""

import os.path
import re

import numpy

from .. import config, utilities
from ..exception import InputError
# relative imports from xrayutilities
from .helper import xu_h5open, xu_open

# define some uesfull regular expressions
SPEC_time_format = re.compile(r"\d\d:\d\d:\d\d")
SPEC_multi_blank = re.compile(r"\s+")
SPEC_multi_blank2 = re.compile(r"\s\s+")
# denotes a numeric value
SPEC_int_value = re.compile(r"[+-]?\d+")
SPEC_num_value = re.compile(
    r"([+-]?\d*\.*\d*[eE]*[+-]*\d+|[+-]?[Ii][Nn][Ff]|[Nn][Aa][Nn])")
SPEC_dataline = re.compile(r"^[+-]*\d.*")

SPEC_scan = re.compile(r"^#S")
SPEC_initmoponames = re.compile(r"#O\d+")
SPEC_initmopopos = re.compile(r"#P\d+")
SPEC_datetime = re.compile(r"^#D")
SPEC_exptime = re.compile(r"^#T")
SPEC_nofcols = re.compile(r"^#N")
SPEC_colnames = re.compile(r"^#L")
SPEC_MCAFormat = re.compile(r"^#@MCA")
SPEC_MCAChannels = re.compile(r"^#@CHANN")
SPEC_headerline = re.compile(r"^#")
SPEC_scanbroken = re.compile(r"#C[a-zA-Z0-9: .]*Scan aborted")
SPEC_scanresumed = re.compile(r"#C[a-zA-Z0-9: .]*Scan resumed")
SPEC_commentline = re.compile(r"#C")
SPEC_newheader = re.compile(r"^#E")
SPEC_errorbm20 = re.compile(r"^MI:")
scan_status_flags = ["OK", "NODATA", "ABORTED", "CORRUPTED"]


class SPECScan(object):
    """
    Represents a single SPEC scan. This class is usually not called by the
    user directly but used via the SPECFile class.
    """

    def __init__(self, name, scannr, command, date, time, itime, colnames,
                 hoffset, doffset, fname, imopnames, imopvalues, scan_status):
        """
        Constructor for the SPECScan class.

        Parameters
        ----------
        name :	    str
            name of the scan
        scannr :    int
            Number of the scan in the specfile
        command :   str
            command used to write the scan
        date :	    str
            starting date of the scan
        time :      str
            starting time of the scan
        itime :	    int
            integration time
        colnames :  list
            list of names of the data columns
        hoffset :   int
            file byte offset to the header of the scan
        doffset :   int
            file byte offset to the data section of the scan
        fname :	    str
            file name of the SPEC file the scan belongs to
        imopnames : list of str
            motor names for the initial motor positions array
        imopvalues : list
            intial motor positions array
        scan_status : {'OK', 'NODATA', 'CORRUPTED', 'ABORTED'}
            scan status as string
        """
        self.name = name  # name of the scan
        self.nr = scannr  # number of the scan
        self.command = command  # command used to record the data
        self.date = date  # date the command has been sent
        self.time = time  # time the command has been sent
        self.colnames = colnames  # list with column names
        self.hoffset = hoffset  # file offset where the header data starts
        self.doffset = doffset  # file offset where the data section starts
        self.fname = fname  # full file name of the file holding the data
        # flag to force resave to hdf5 file in Save2HDF5()
        self.ischanged = True
        self.header = []
        self.fid = None

        if scan_status in scan_status_flags:
            self.scan_status = scan_status
        else:
            self.scan_status = "CORRUPTED"
            if config.VERBOSITY >= config.INFO_ALL:
                print("XU.io.spec.SPECScan: unknown scan status flag - "
                      "set to CORRUPTED")

        # setup the initial motor positions dictionary - set the motor names
        # dictionary holding the initial motor positions
        self.init_motor_pos = {}
        if len(imopnames) == len(imopvalues):
            for i in range(len(imopnames)):
                natmotname = utilities.makeNaturalName(imopnames[i])
                self.init_motor_pos[
                    "INIT_MOPO_" + natmotname] = float(imopvalues[i])
        else:
            print("XU.io.spec.SPECScan: Warning: incorrect number of "
                  "initial motor positions in scan %03d" % (self.nr))
            if config.VERBOSITY >= config.INFO_ALL:
                print(imopnames)
                print(imopvalues)
            # ASSUME ORDER DID NOT CHANGE!! (which might be wrong)
            # in fact this is sign for a broken spec file
            # number of initial motor positions should not change without new
            # file header!
            for i in range(min(len(imopnames), len(imopvalues))):
                natmotname = utilities.makeNaturalName(imopnames[i])
                self.init_motor_pos[
                    "INIT_MOPO_" + natmotname] = float(imopvalues[i])
            # read the rest of the positions into dummy INIT_MOPO__NONAME__%03d
            for i in range(len(imopnames), len(imopvalues)):
                self.init_motor_pos["INIT_MOPO___NONAME__%03d" % (i)] = \
                    float(imopvalues[i])

        # some additional attributes for the MCA data
        # False if scan contains no MCA data, True otherwise
        self.has_mca = False
        self.mca_column_format = 0  # number of columns used to save MCA data
        self.mca_channels = 0  # number of channels stored from the MCA
        self.mca_nof_lines = 0  # number of lines used to store MCA data
        self.mca_start_channel = 0  # first channel of the MCA that is stored
        self.mca_stop_channel = 0  # last channel of the MCA that is stored

        # a numpy record array holding the data - this is set using by the
        # ReadData method.
        self.data = None

        # check for duplicate values in column names
        for i in range(len(self.colnames)):
            name = self.colnames[i]
            cnt = self.colnames.count(name)
            if cnt > 1:
                # have multiple entries
                cnt = 1
                for j in range(self.colnames.index(name) + 1,
                               len(self.colnames)):
                    if self.colnames[j] == name:
                        self.colnames[j] = name + "_%i" % cnt
                    cnt += 1

    def SetMCAParams(self, mca_column_format, mca_channels,
                     mca_start, mca_stop):
        """
        Set the parameters used to save the MCA data to the file. This method
        calculates the number of lines used to store the MCA data from the
        number of columns and the

        Parameters
        ----------
        mca_column_format : int
            number of columns used to save the data
        mca_channels :	    int
            number of MCA channels stored
        mca_start :	    int
            first channel that is stored
        mca_stop :          int
            last channel that is stored
        """
        self.has_mca = True
        self.mca_column_format = mca_column_format
        self.mca_channels = mca_channels
        self.mca_start_channel = mca_start
        self.mca_stop_channel = mca_stop

        # calculate the number of lines per data point for the mca
        self.mca_nof_lines = int(mca_channels / mca_column_format)
        if mca_channels % mca_column_format != 0:
            # some additional values have to be read
            self.mca_nof_lines = self.mca_nof_lines + 1

        if config.VERBOSITY >= config.DEBUG:
            print("XU.io.SPECScan.SetMCAParams: number of channels: %d"
                  % self.mca_channels)
            print("XU.io.SPECScan.SetMCAParams: number of columns: %d"
                  % self.mca_column_format)
            print("XU.io.SPECScan.SetMCAParams: number of lines to read "
                  "for MCA: %d" % self.mca_nof_lines)

    def __str__(self):
        # build a proper string to print the scan information
        str_rep = "|%4i|" % (self.nr)
        str_rep = str_rep + "%50s|%10s|%10s|" % (self.command, self.time,
                                                 self.date)

        if self.has_mca:
            str_rep = str_rep + "MCA: %5i" % (self.mca_channels)

        str_rep = str_rep + "\n"
        return str_rep

    def ClearData(self):
        """
        Delete the data stored in a scan after it is no longer
        used.
        """

        self.__delattr__("data")
        self.data = None

    def ReadData(self):
        """
        Set the data attribute of the scan class.
        """

        if self.scan_status == "NODATA":
            if config.VERBOSITY >= config.INFO_LOW:
                print("XU.io.SPECScan.ReadData: %s has been aborted - "
                      "no data available!" % self.name)
            self.data = None
            return None

        if not self.has_mca:
            if config.VERBOSITY >= config.INFO_ALL:
                print("XU.io.SPECScan.ReadData: scan %d contains no MCA data"
                      % self.nr)

        with xu_open(self.fname) as self.fid:
            # read header lines
            self.fid.seek(self.hoffset, 0)
            self.header = []
            while self.fid.tell() < self.doffset:
                line = self.fid.readline().decode('ascii', 'ignore')
                self.header.append(line.strip())

            self.fid.seek(self.doffset, 0)

            # create dictionary to hold the data
            if self.has_mca:
                type_desc = {"names": self.colnames + ["MCA"],
                             "formats": len(self.colnames) * [numpy.float32] +
                             [(numpy.uint32, self.mca_channels)]}
            else:
                type_desc = {"names": self.colnames,
                             "formats": len(self.colnames) * [numpy.float32]}

            if config.VERBOSITY >= config.DEBUG:
                print("xu.io.SPECScan.ReadData: type descriptor: %s"
                      % (repr(type_desc)))

            record_list = []  # from this list the record array while be built

            mca_counter = 0
            scan_aborted_flag = False

            for line in self.fid:
                line = line.decode('ascii', 'ignore')
                line = line.strip()
                if not line:
                    continue

                # check if scan is broken
                if (SPEC_scanbroken.findall(line) != [] or
                        scan_aborted_flag):
                    # need to check next line(s) to know if scan is resumed
                    # read until end of comment block or end of file
                    if not scan_aborted_flag:
                        scan_aborted_flag = True
                        self.scan_status = "ABORTED"
                        if config.VERBOSITY >= config.INFO_ALL:
                            print("XU.io.SPECScan.ReadData: %s aborted"
                                  % self.name)
                        continue
                    elif SPEC_scanresumed.match(line):
                        self.scan_status = "OK"
                        scan_aborted_flag = False
                        if config.VERBOSITY >= config.INFO_ALL:
                            print("XU.io.SPECScan.ReadData: %s resumed"
                                  % self.name)
                        continue
                    elif SPEC_commentline.match(line):
                        continue
                    elif SPEC_errorbm20.match(line):
                        print(line)
                        continue
                    else:
                        break

                if SPEC_headerline.match(line) or \
                   SPEC_commentline.match(line):
                    if SPEC_scanresumed.match(line):
                        continue
                    elif SPEC_commentline.match(line):
                        continue
                    else:
                        break

                if mca_counter == 0:
                    # the line is a scalar data line
                    line_list = SPEC_num_value.findall(line)
                    if config.VERBOSITY >= config.DEBUG:
                        print("XU.io.SPECScan.ReadData: %s" % line)
                        print("XU.io.SPECScan.ReadData: read scalar values %s"
                              % repr(line_list))
                    # convert strings to numbers
                    line_list = map(float, line_list)

                    # increment the MCA counter if MCA data is stored
                    if self.has_mca:
                        mca_counter = mca_counter + 1
                        # create a temporary list for the mca data
                        mca_tmp_list = []
                    else:
                        record_list.append(tuple(line_list))
                else:
                    # reading MCA spectrum
                    mca_tmp_list += map(int, SPEC_int_value.findall(line))

                    # increment MCA counter
                    mca_counter = mca_counter + 1
                    # if mca_counter exceeds the number of lines used to store
                    # MCA data: append everything to the record list
                    if mca_counter > self.mca_nof_lines:
                        record_list.append(tuple(list(line_list) +
                                                 [mca_tmp_list]))
                        mca_counter = 0

            # convert the data to numpy arrays
            ncol = len(record_list[0])
            if config.VERBOSITY >= config.INFO_LOW:
                print("XU.io.SPECScan.ReadData: %s: %d %d %d"
                      % (self.name, len(record_list), ncol,
                         len(type_desc["names"])))
            if ncol == len(type_desc["names"]):
                try:
                    self.data = numpy.rec.fromrecords(record_list,
                                                      dtype=type_desc)
                except ValueError:
                    self.scan_status = 'NODATA'
                    print("XU.io.SPECScan.ReadData: %s exception while "
                          "parsing data" % self.name)
            else:
                self.scan_status = 'NODATA'

    def plot(self, *args, **keyargs):
        """
        Plot scan data to a matplotlib figure. If newfig=True a new
        figure instance will be created. If logy=True (default is False)
        the y-axis will be plotted with a logarithmic scale.

        Parameters
        ----------
        args :  list
            arguments for the plot: first argument is the name of x-value
            column the following pairs of arguments are the y-value names and
            plot styles allowed are 3, 5, 7,... number of arguments
        keyargs :   dict, optional
        newfig :    bool, optional
            if True a new figure instance will be created otherwise an existing
            one will be used
        logy :      bool, optional
            if True a semilogy plot will be done
        """
        flag, plt = utilities.import_matplotlib_pyplot('XU.io.SPECScan')
        if not flag:
            return

        newfig = keyargs.get('newfig', True)
        logy = keyargs.get('logy', False)

        try:
            xname = args[0]
            xdata = self.data[xname]
        except ValueError:
            raise InputError("name of the x-axis is invalid!")

        alist = args[1:]
        leglist = []

        if len(alist) % 2 != 0:
            raise InputError("wrong number of yname/style arguments!")

        if newfig:
            plt.figure()
            plt.subplots_adjust(left=0.08, right=0.95)

        for i in range(0, len(alist), 2):
            yname = alist[i]
            ystyle = alist[i + 1]
            try:
                ydata = self.data[yname]
            except ValueError:
                raise InputError("no column with name %s exists!" % yname)
                continue
            if logy:
                plt.semilogy(xdata, ydata, ystyle)
            else:
                plt.plot(xdata, ydata, ystyle)

            leglist.append(yname)

        plt.xlabel("%s" % xname)
        plt.legend(leglist)
        plt.title("scan %i %s\n%s %s"
                  % (self.nr, self.command, self.date, self.time))
        # need to adjust axis limits properly
        lim = plt.axis()
        plt.axis([xdata.min(), xdata.max(), lim[2], lim[3]])

    def Save2HDF5(self, h5f, group="/", title="", optattrs={}, comp=True):
        """
        Save a SPEC scan to an HDF5 file. The method creates a group with the
        name of the scan and stores the data there as a table object with name
        "data". By default the scan group is created under the root group of
        the HDF5 file.  The title of the scan group is ususally the scan
        command. Metadata of the scan are stored as attributes to the scan
        group. Additional custom attributes to the scan group can be passed as
        a dictionary via the optattrs keyword argument.

        Parameters
        ----------
        h5f :	 file-handle or str
            a HDF5 file object or its filename

        group :	 str, optional
            name or group object of the HDF5 group where to store the data
        title :	 str, optional
            a string with the title for the data, defaults to the name of scan
            if empty
        optattrs : dict, optional
            a dictionary with optional attributes to store for the data
        comp :	 bool, optional
            activate compression - true by default
        """

        with xu_h5open(h5f, 'a') as h5:
            # check if data object has been already written
            if self.data is None:
                raise InputError("XU.io.SPECScan.Save2HDF5: No data has been"
                                 "read so far - call ReadData method of the "
                                 "scan")
                return None

            # parse keyword arguments:
            if isinstance(group, str):
                rootgroup = h5.get(group)
            else:
                rootgroup = group

            if title != "":
                group_title = title
            else:
                group_title = self.name
            group_title = group_title.replace(".", "_")

            # create the dataset and fill it
            copy_count = 0
            if self.ischanged and group_title in rootgroup:
                del rootgroup[group_title]
            raw_grp_title = group_title
            # if the group already exists the name must be changed and
            # another will be made to create the group.
            while group_title in rootgroup:
                group_title = raw_grp_title + "_%i" % (copy_count)
                copy_count = copy_count + 1
            g = rootgroup.create_group(group_title)

            kwds = {'fletcher32': True}
            if comp:
                kwds['compression'] = 'gzip'

            g.create_dataset("data", data=self.data, **kwds)

            # write attribute data for the scan
            g.attrs['ScanNumber'] = numpy.uint(self.nr)
            g.attrs['Command'] = self.command
            g.attrs['Date'] = self.date
            g.attrs['Time'] = self.time
            g.attrs['scan_status'] = self.scan_status

            # write the initial motor positions as attributes
            for k in self.init_motor_pos:
                g.attrs[k] = numpy.float(self.init_motor_pos[k])

            # if scan contains MCA data write also MCA parameters
            g.attrs['has_mca'] = self.has_mca
            g.attrs['mca_start_channel'] = numpy.uint(self.mca_start_channel)
            g.attrs['mca_stop_channel'] = numpy.uint(self.mca_stop_channel)
            g.attrs['mca_nof_channels'] = numpy.uint(self.mca_channels)

            for k in optattrs:
                g.attrs[k] = optattrs[k]

            h5.flush()

    def getheader_element(self, key, firstonly=True):
        """
        return the value-string of the first appearance of this SPECScan's
        header element, or a list of all values if firstonly=False

        Parameters
        ----------
        specscan :  SPECScan
        key :       str
            name of the key to return; e.g. 'UMONO' or 'D'
        firstonly : bool, optional
            flag to specify if all instances or only the first one should be
            returned

        Returns
        -------
        valuestring :   str
            header value (if firstonly=True)
        [str1, str2, ...] : list
            header values (if firstonly=False)
        """
        if not self.header:
            self.ReadData()
        re_key = re.compile(r'^#%s (.*)' % key)
        ret = []
        for line in self.header:
            m = re_key.match(line)
            if m:
                if firstonly:
                    ret = m.groups()[0]
                    break
                else:
                    ret.append(m.groups()[0])
        return ret


class SPECFile(object):

    """
    This class represents a single SPEC file. The class provides
    methodes for updateing an already opened file which makes it particular
    interesting for interactive use.
    """

    def __init__(self, filename, path=""):
        """
        SPECFile init routine

        Parameters
        ----------
        filename :  str
            filename of the spec file
        path :      str, optional
            path to the specfile
        """
        self.full_filename = os.path.join(path, filename)
        self.filename = os.path.basename(self.full_filename)

        # list holding scan objects
        self.scan_list = []
        self.fid = None
        self.last_offset = 0

        # initially parse the file
        self.init_motor_names_fh = []  # this list will hold the names of the
        # motors saved in initial motor positions given in the file header
        self.init_motor_names_sh = []  # this list will hold the names of the
        # motors saved in initial motor positions given in the scan header
        self.init_motor_names = []  # this list will hold the names of the
        # motors saved in initial motor positions from either the file or
        # scan header

        self.Parse()

    def __getitem__(self, index):
        """
        function to return the n-th scan in the spec-file.  be aware that
        numbering starts at 0! If scans are missing the relation between the
        given number and the "number" of the returned scan might be not
        trivial.

        See also
        --------
        scanI
            attributes of the SPECFile object, where 'I' is the scan number
        """
        return self.scan_list[index]

    def __getattr__(self, name):
        """
        return scanX objects where X stands for the scan number in the SPECFile
        which for this purpose is assumed to be unique. (otherwise the first
        instance of scan number X is returned)
        """
        if name.startswith("scan"):
            index = name[4:]

            try:
                scannr = int(index)
            except ValueError:
                raise AttributeError("scannumber needs to be convertable to "
                                     "integer")

            # try to find the scan in the list of scans
            s = None
            for scan in self.scan_list:
                if scan.nr == scannr:
                    s = scan
                    break

            if s is not None:
                return s
            else:
                raise AttributeError("requested scan-number not found")
        else:
            raise AttributeError("SPECFile has no attribute '%s'" % name)

    def __len__(self):
        return self.scan_list.__len__()

    def __str__(self):
        ostr = ""
        for i in range(len(self.scan_list)):
            ostr = ostr + "%5i" % (i)
            ostr = ostr + self.scan_list[i].__str__()

        return ostr

    def Save2HDF5(self, h5f, comp=True, optattrs={}):
        """
        Save the entire file in an HDF5 file. For that purpose a group is set
        up in the root group of the file with the name of the file without
        extension and leading path.  If the method is called after an previous
        update only the scans not written to the file meanwhile are saved.

        Parameters
        ----------
        h5f :   file-handle or str
            a HDF5 file object or its filename
        comp :  bool, optional
            activate compression - true by default
        """
        with xu_h5open(h5f, 'a') as h5:
            groupname = os.path.splitext(os.path.splitext(self.filename)[0])[0]
            try:
                g = h5.create_group(groupname)
            except ValueError:
                g = h5.get(groupname)

            g.attrs['TITLE'] = "Data of SPEC - File %s" % (self.filename)
            for k in optattrs:
                g.attrs[k] = optattrs[k]
            for s in self.scan_list:
                if (((s.name not in g) or s.ischanged) and
                        s.scan_status != "NODATA"):
                    s.ReadData()
                    if s.data is not None:
                        s.Save2HDF5(h5, group=g, comp=comp)
                        s.ClearData()
                        s.ischanged = False

    def Update(self):
        """
        reread the file and add newly added files. The parsing starts at the
        data offset of the last scan gathered during the last parsing run.
        """

        # reparse the SPEC file
        if config.VERBOSITY >= config.INFO_LOW:
            print("XU.io.SPECFile.Update: reparsing file for new scans ...")
        # mark last found scan as not saved to force reread
        idx = len(self.scan_list)
        if idx > 0:
            lastscan = self.scan_list[idx - 1]
            lastscan.ischanged = True
        self.Parse()

    def Parse(self):
        """
        Parses the file from the starting at last_offset and adding found scans
        to the scan list.
        """
        with xu_open(self.full_filename) as self.fid:
            # move to the last read position in the file
            self.fid.seek(self.last_offset, 0)
            scan_started = False
            scan_has_mca = False
            # list with the motors from whome the initial
            # position is stored.
            init_motor_values = []

            if config.VERBOSITY >= config.DEBUG:
                print('XU.io.SPECFile: start parsing')

            for line in self.fid:
                linelength = len(line)
                line = line.decode('ascii', 'ignore')
                if config.VERBOSITY >= config.DEBUG:
                    print('parsing line: %s' % line)

                # remove trailing and leading blanks from the read line
                line = line.strip()

                # fill the list with the initial motor names in the header
                if SPEC_newheader.match(line):
                    self.init_motor_names_fh = []

                elif SPEC_initmoponames.match(line) and not scan_started:
                    if config.VERBOSITY >= config.DEBUG:
                        print("XU.io.SPECFile.Parse: found initial motor "
                              "names in file header")
                    line = SPEC_initmoponames.sub("", line)
                    line = line.strip()
                    self.init_motor_names_fh = self.init_motor_names_fh + \
                        SPEC_multi_blank2.split(line)

                # if the line marks the beginning of a new scan
                elif SPEC_scan.match(line) and not scan_started:
                    if config.VERBOSITY >= config.DEBUG:
                        print("XU.io.SPECFile.Parse: found scan")
                    line_list = SPEC_multi_blank.split(line)
                    scannr = int(line_list[1])
                    scancmd = "".join(" " + x + " " for x in line_list[2:])
                    scan_started = True
                    scan_has_mca = False
                    scan_header_offset = self.last_offset
                    scan_status = "OK"
                    # define some necessary variables which could be missing in
                    # the scan header
                    itime = numpy.nan
                    time = ''
                    if config.VERBOSITY >= config.INFO_ALL:
                        print("XU.io.SPECFile.Parse: processing scan nr. %d "
                              "..." % scannr)
                    # set the init_motor_names to the ones found in
                    # the file header
                    self.init_motor_names_sh = []
                    self.init_motor_names = self.init_motor_names_fh

                    # if the line contains the date and time information
                elif SPEC_datetime.match(line) and scan_started:
                    if config.VERBOSITY >= config.DEBUG:
                        print("XU.io.SPECFile.Parse: found date and time")
                    # fetch the time from the line data
                    time = SPEC_time_format.findall(line)[0]
                    line = SPEC_time_format.sub("", line)
                    line = SPEC_datetime.sub("", line)
                    date = SPEC_multi_blank.sub(" ", line).strip()

                # if the line contains the integration time
                elif SPEC_exptime.match(line) and scan_started:
                    if config.VERBOSITY >= config.DEBUG:
                        print("XU.io.SPECFile.Parse: found exposure time")
                    itime = float(SPEC_num_value.findall(line)[0])
                # read the initial motor names in the scan header if present
                elif SPEC_initmoponames.match(line) and scan_started:
                    if config.VERBOSITY >= config.DEBUG:
                        print("XU.io.SPECFile.Parse: found initial motor "
                              "names in scan header")
                    line = SPEC_initmoponames.sub("", line)
                    line = line.strip()
                    self.init_motor_names_sh = self.init_motor_names_sh + \
                        SPEC_multi_blank2.split(line)
                    self.init_motor_names = self.init_motor_names_sh
                # read the initial motor positions
                elif SPEC_initmopopos.match(line) and scan_started:
                    if config.VERBOSITY >= config.DEBUG:
                        print("XU.io.SPECFile.Parse: found initial motor "
                              "positions")
                    line = SPEC_initmopopos.sub("", line)
                    line = line.strip()
                    line_list = SPEC_multi_blank.split(line)
                    # sometimes initial motor position are simply empty and
                    # this should not lead to an error
                    try:
                        for value in line_list:
                            init_motor_values.append(float(value))
                    except ValueError:
                        pass

                # if the line contains the number of colunmns
                elif SPEC_nofcols.match(line) and scan_started:
                    if config.VERBOSITY >= config.DEBUG:
                        print("XU.io.SPECFile.Parse: found number of columns")
                    line = SPEC_nofcols.sub("", line)
                    line = line.strip()
                    nofcols = int(line)

                # if the line contains the column names
                elif SPEC_colnames.match(line) and scan_started:
                    if config.VERBOSITY >= config.DEBUG:
                        print("XU.io.SPECFile.Parse: found column names")
                    line = SPEC_colnames.sub("", line)
                    line = line.strip()
                    col_names = SPEC_multi_blank.split(line)

                    # this is a fix in the case that blanks are allowed in
                    # motor and detector names (only a single balanks is
                    # supported meanwhile)
                    if len(col_names) > nofcols:
                        col_names = SPEC_multi_blank2.split(line)

                elif SPEC_MCAFormat.match(line) and scan_started:
                    mca_col_number = int(SPEC_num_value.findall(
                                         line)[0])
                    scan_has_mca = True

                elif SPEC_MCAChannels.match(line) and scan_started:
                    line_list = SPEC_num_value.findall(line)
                    mca_channels = int(line_list[0])
                    mca_start = int(line_list[1])
                    mca_stop = int(line_list[2])

                elif (SPEC_scanbroken.findall(line) != [] and
                      scan_started):
                    # this is the case when a scan is broken and no data has
                    # been written, but nevertheless a comment is in the file
                    # that tells us that the scan was aborted
                    scan_data_offset = self.last_offset
                    s = SPECScan("scan_%i" % (scannr), scannr, scancmd,
                                 date, time, itime, col_names,
                                 scan_header_offset, scan_data_offset,
                                 self.full_filename, self.init_motor_names,
                                 init_motor_values, "NODATA")

                    self.scan_list.append(s)

                    # reset control flags
                    scan_started = False
                    scan_has_mca = False
                    # reset initial motor positions flag
                    init_motor_values = []

                elif SPEC_dataline.match(line) and scan_started:
                    # this is now the real end of the header block. at this
                    # point we know that there is enough information about the
                    # scan

                    # save the data offset
                    scan_data_offset = self.last_offset

                    # create an SPECFile scan object and add it to the scan
                    # list the name of the group consists of the prefix scan
                    # and the number of the scan in the file - this shoule make
                    # it easier to find scans in the HDF5 file.
                    s = SPECScan("scan_%i" % (scannr), scannr, scancmd, date,
                                 time, itime, col_names, scan_header_offset,
                                 scan_data_offset, self.full_filename,
                                 self.init_motor_names, init_motor_values,
                                 scan_status)
                    if scan_has_mca:
                        s.SetMCAParams(mca_col_number, mca_channels, mca_start,
                                       mca_stop)

                    self.scan_list.append(s)

                    # reset control flags
                    scan_started = False
                    scan_has_mca = False
                    # reset initial motor positions flag
                    init_motor_values = []

                elif SPEC_scan.match(line) and scan_started:
                    # this should only be the case when there are two
                    # consecutive file headers in the data file without any
                    # data or abort notice of the first scan; first store
                    # current scan as aborted then start new scan parsing
                    s = SPECScan("scan_%i" % (scannr), scannr, scancmd,
                                 date, time, itime, col_names,
                                 scan_header_offset, None,
                                 self.full_filename, self.init_motor_names,
                                 init_motor_values, "NODATA")
                    self.scan_list.append(s)

                    # reset control flags
                    scan_started = False
                    scan_has_mca = False
                    # reset initial motor positions flag
                    init_motor_values = []

                    # start parsing of new scan
                    if config.VERBOSITY >= config.DEBUG:
                        print("XU.io.SPECFile.Parse: found scan "
                              "(after aborted scan)")
                    line_list = SPEC_multi_blank.split(line)
                    scannr = int(line_list[1])
                    scancmd = "".join(" " + x + " " for x in line_list[2:])
                    scan_started = True
                    scan_has_mca = False
                    scan_header_offset = self.last_offset
                    scan_status = "OK"
                    self.init_motor_names_sh = []
                    self.init_motor_names = self.init_motor_names_fh

                # store the position of the file pointer
                self.last_offset += linelength

            # if reading of the file is finished store the data offset of the
            # last scan as the last offset for the next parsing run of the file
            self.last_offset = self.scan_list[-1].doffset


class SPECCmdLine(object):

    def __init__(self, n, prompt, cmdl, out=""):
        self.linenumber = n
        self.prompt = prompt
        self.command = cmdl
        self.out = out

    def __str__(self):
        ostr = "%i.%s> %s" % (self.linenumber, self.prompt, self.command)
        return ostr


class SPECLog(object):
    """
    class to parse a SPEC log file to find the command history
    """

    def __init__(self, filename, prompt, path=""):
        """
        init routine for a class to read a SPEC log file

        Parameters
        ----------
        filename :  str
            SPEC log file name
        prompt :    str
            SPEC command prompt (e.g. 'PSIC' or 'SPEC')
        path :      str, optional
            directory where the SPEC log can be found
        """
        self.filename = filename
        self.full_filename = os.path.join(path, self.filename)

        self.prompt = prompt
        self.prompt_re = re.compile(r"%s>" % self.prompt)

        self.cmdl_list = []
        self.line_counter = 0
        self.Parse()

    def Parse(self):
        with xu_open(self.full_filename, 'r') as fid:
            for line in fid:
                line = line.decode('ascii', 'ignore')
                self.line_counter += 1

                line = line.strip()
                if self.prompt_re.findall(line):
                    [line, cmd] = self.prompt_re.split(line)
                    self.cmdl_list.append(SPECCmdLine(int(float(line)),
                                                      self.prompt, cmd))

    def __getitem__(self, index):
        """
        function to return the n-th cmd in the spec-log.
        """
        return self.cmdl_list[index]

    def __str__(self):
        ostr = "%s with %d lines\n" % (self.filename, self.line_counter)

        for cmd in self.cmdl_list:
            ostr = ostr + cmd.__str__() + "\n"

        return ostr


def geth5_scan(h5f, scans, *args, **kwargs):
    """
    function to obtain the angular cooridinates as well as intensity values
    saved in an HDF5 file, which was created from a spec file by the Save2HDF5
    method. Especially useful for reciprocal space map measurements.

    further more it is possible to obtain even more positions from
    the data file if more than two string arguments with its names are given

    Parameters
    ----------
    h5f :       file-handle or str
        file object of a HDF5 file opened using h5py or its filename
    scans :     int, tuple or list
        number of the scans of the reciprocal space map
    args :      str, optional
        names of the motors. to read reciprocal space maps measured in coplanar
        diffraction give:

         - omname: name of the omega motor (or its equivalent)
         - ttname: name of the two theta motor (or its equivalent)

    kwargs :    dict, optional
    samplename: str, optional
        string with the hdf5-group containing the scan data if ommited the
        first child node of h5f.root will be used
    rettype:    {'list', 'numpy'}, optional
        how to return motor positions. by default a list of arrays is returned.
        when rettype == 'numpy' a record array will be returned.

    Returns
    -------
    [ang1, ang2, ...] :     list
        angular positions of the center channel of the position sensitive
        detector (numpy.ndarray 1D), this list is omitted if no `args` are
        given
    MAP :   ndarray
        the data values as stored in the data file (includes the intensities
        e.g. MAP['MCA']).

    Examples
    --------
    >>> [om, tt], MAP = xu.io.geth5_scan(h5file, 36, 'omega', 'gamma')
    """

    with xu_h5open(h5f) as h5:
        gname = kwargs.get("samplename", list(h5.keys())[0])
        h5g = h5.get(gname)

        if numpy.iterable(scans):
            scanlist = scans
        else:
            scanlist = list([scans])

        angles = dict.fromkeys(args)
        for key in angles:
            if not isinstance(key, str):
                raise InputError("*arg values need to be strings with "
                                 "motornames")
            angles[key] = numpy.zeros(0)
        buf = numpy.zeros(0)
        MAP = numpy.zeros(0)

        for nr in scanlist:
            h5scan = h5g.get("scan_%d" % nr)
            sdata = numpy.asarray(h5scan.get('data'))
            if MAP.dtype == numpy.float64:
                MAP.dtype = sdata.dtype
            # append scan data to MAP, where all data are stored
            MAP = numpy.append(MAP, sdata)
            # check type of scan
            notscanmotors = []
            for i in range(len(args)):
                motname = args[i]
                try:
                    buf = sdata[motname]
                    scanshape = buf.shape
                    angles[motname] = numpy.concatenate((angles[motname], buf))
                except ValueError:
                    notscanmotors.append(i)
            if len(notscanmotors) == len(args):
                scanshape = len(sdata)
            for i in notscanmotors:
                motname = args[i]
                natmotname = utilities.makeNaturalName(motname)
                buf = numpy.ones(scanshape) * \
                    h5scan.attrs["INIT_MOPO_%s" % natmotname]
                angles[motname] = numpy.concatenate((angles[motname], buf))

    # create return values in correct order
    def create_retval():
        retval = []
        for motname in args:
            retval.append(angles[motname])
        return retval

    rettype = kwargs.get('rettype', 'list')
    if rettype == 'numpy':
        retval = numpy.core.records.fromarrays([angles[m] for m in args],
                                               names=args)
    else:
        retval = create_retval()

    if not args:
        return MAP
    else:
        return retval, MAP


def getspec_scan(specf, scans, *args, **kwargs):
    """
    function to obtain the angular cooridinates as well as intensity values
    saved in a SPECFile. Especially useful to combine the data from multiple
    scans.

    further more it is possible to obtain even more positions from
    the data file if more than two string arguments with its names are given

    Parameters
    ----------
    specf :     SPECFile
        file object
    scans :     int, tuple or list
        number of the scans
    args :      str
        names of the motors and counters
    rettype :   {'list', 'numpy'}, optional
        how to return motor positions. by default a list of arrays is returned.
        when rettype == 'numpy' a record array will be returned.

    Returns
    -------
    [ang1, ang2, ...] : list
        coordinates and counters from the SPEC file

    Examples
    --------
    >>> [om, tt, cnt2] = xu.io.getspec_scan(s, 36, 'omega', 'gamma',
    >>>                                     'Counter2')
    """
    if not args:
        return

    if numpy.iterable(scans):
        scanlist = scans
    else:
        scanlist = list([scans])

    angles = dict.fromkeys(args)
    for key in angles:
        if not isinstance(key, str):
            raise InputError("*arg values need to be strings with "
                             "motornames")
        angles[key] = numpy.zeros(0)
        buf = numpy.zeros(0)

    for nr in scanlist:
        sscan = specf.__getattr__("scan%d" % nr)
        sscan.ReadData()
        sdata = sscan.data
        # check type of scan
        notscanmotors = []
        for i in range(len(args)):
            motname = args[i]
            try:
                buf = sdata[motname]
                scanshape = buf.shape
                angles[motname] = numpy.concatenate((angles[motname], buf))
            except ValueError:
                notscanmotors.append(i)
        if len(notscanmotors) == len(args):
            scanshape = len(sdata)
        for i in notscanmotors:
            motname = args[i]
            buf = (numpy.ones(scanshape) *
                   sscan.init_motor_pos["INIT_MOPO_%s"
                                        % utilities.makeNaturalName(motname)])
            angles[motname] = numpy.concatenate((angles[motname], buf))

    # create return values in correct order
    def create_retval():
        retval = []
        for motname in args:
            retval.append(angles[motname])
        return retval

    rettype = kwargs.get('rettype', 'list')
    if rettype == 'numpy':
        retval = numpy.core.records.fromarrays([angles[m] for m in args],
                                               names=args)
    else:
        retval = create_retval()

    return retval



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2020 Dominik Kriegner <dominik.kriegner@gmail.com>

import math
import os
import unittest

import numpy
import xrayutilities as xu

xu.config.VERBOSITY = 0


@unittest.skipIf('CI' in os.environ, "slow test not running on CI")
class Test_Materials_reflection_condition(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.p = xu.PowderExperiment(en=10000)
        cls.ksinmax = cls.p.k0 * math.sin(math.radians(179.9)/2)
        cls.materials = []
        for name, obj in xu.materials.predefined_materials.__dict__.items():
            if isinstance(obj, xu.materials.Crystal):
                cls.materials.append(obj)

    def _test_material(self, m, test_allowed=False):
        """
        helper function to test one specific material

        Parameters
        ----------
        m: xu.materials.Crystal
         material to be tested
        test_allowed: bool
         boolean to determine wether allowed peaks must have non-zero structure
         factor. This test is only performed if reflection conditions are
         available, and two atoms have a minimal distance of 0.01Ang! still
         there is a chance that by chance atoms occupy positions of a
         different space group which renders this test option dangerous.
         default: False
        """
        # calculate maximal Bragg indices
        hma = int(math.ceil(m.a / math.pi * self.ksinmax))
        hmi = -hma
        kma = int(math.ceil(m.b / math.pi * self.ksinmax))
        kmi = -kma
        lma = int(math.ceil(m.c / math.pi * self.ksinmax))
        lmi = -lma

        # calculate structure factors
        qmax = 2 * self.ksinmax
        hkl = numpy.mgrid[hma:hmi-1:-1,
                          kma:kmi-1:-1,
                          lma:lmi-1:-1].reshape(3, -1).T

        q = m.Q(hkl)
        qnorm = numpy.linalg.norm(q, axis=1)
        mask = numpy.logical_and(qnorm > 0, qnorm <= qmax)
        errorinfo = str(m)
        allowed = numpy.asarray([m.lattice.hkl_allowed(b) for b in hkl[mask]])
        # all forbidden peaks must have structure factor close to zero
        s = numpy.abs(
            m.StructureFactorForQ(q[mask][numpy.logical_not(allowed)],
                                  self.p.energy))
        mviolate = numpy.logical_not(numpy.isclose(s, 0))
        if numpy.any(mviolate):
            for h, sf in zip(hkl[mask][numpy.logical_not(allowed)][mviolate],
                             s[mviolate]):
                errorinfo += "%s\t%s\n" % (h, sf)
        self.assertTrue(numpy.allclose(s, 0), msg=errorinfo)
        # check if atoms are too near (reduce chance of accidental extinction)
        if test_allowed:
            for at, pos, occ, b in m.lattice.base():
                env = m.environment(*pos, maxdist=0.01)
                if len(env) > 1 or env[0][-1] != 1:
                    test_allowed = False
                    break
        if test_allowed and 'n/a' not in m.lattice.reflection_conditions():
            # all allowed peaks must have non-zero structure factor
            s = numpy.abs(
                m.StructureFactorForQ(q[mask][allowed], self.p.energy))
            mviolate = numpy.isclose(s, 0)
            if numpy.any(mviolate):
                for h, sf in zip(hkl[mask][allowed][mviolate], s[mviolate]):
                    errorinfo += "%s\t%s\n" % (h, sf)
            self.assertTrue(numpy.all(numpy.logical_not(numpy.isclose(s, 0))),
                            msg=errorinfo)

    def test_hklcond_predefined(self):
        for m in self.materials:
            self._test_material(m)

    def test_hklcond_random(self):
        # create random materials with random unit cell and test them
        x, y, z = numpy.random.rand(3)
        a, b, c = numpy.random.rand(3) * 2 + 4
        al, be, gam = numpy.random.rand(3) * 60 + 60
        pdict = {'a': a, 'b': b, 'c': c, 'alpha': al, 'beta': be, 'gamma': gam}
        wp = xu.materials.spacegrouplattice.wp

        for sg in wp.keys():
            # determine parameters for this space group
            sgnr = int(sg.split(':')[0])
            csys, nargs = xu.materials.spacegrouplattice.sgrp_sym[sgnr]
            params = xu.materials.spacegrouplattice.sgrp_params[csys][0]
            p = [eval(par, pdict) for par in params]
            # test all Wyckoff positions
            for wplabel in wp[sg].keys():
                wpentry = wp[sg][wplabel]
                wppar = []
                if wpentry[0] & 1:
                    wppar.append(x)
                if wpentry[0] & 2:
                    wppar.append(y)
                if wpentry[0] & 4:
                    wppar.append(z)
                kwdict = {'atoms': [xu.materials.elements.Dummy, ], 'pos':
                          [(wplabel, wppar), ]}
                # generate test lattice
                lat = xu.materials.SGLattice(sg, *p, **kwdict)
                self._test_material(xu.materials.Crystal("SG%s" % sg, lat))

    def test_get_allowed_hkl(self):
        """
        use some random hkls to test the get_allowed_hkl function. The internal
        methods used by this function is also tested more thoroughly by other
        unit tests.
        """
        qmax = 2 * self.ksinmax
        N = 5
        for m in self.materials:
            if 'n/a' in m.lattice.reflection_conditions():
                continue
            hkls = m.lattice.get_allowed_hkl(qmax)
            hma = int(math.ceil(m.a / math.pi * self.ksinmax))
            hmi = -hma
            kma = int(math.ceil(m.b / math.pi * self.ksinmax))
            kmi = -kma
            lma = int(math.ceil(m.c / math.pi * self.ksinmax))
            lmi = -lma
            errorinfo = str(m)
            errorinfo += "HKL min/max: %d %d %d / %d %d %d" % (hmi, kmi, lmi,
                                                               hma, kma, lma)
            for h, k, l in zip(numpy.random.randint(hmi, hma, N),
                               numpy.random.randint(kmi, kma, N),
                               numpy.random.randint(lmi, lma, N)):
                qnorm = numpy.linalg.norm(m.Q(h, k, l))
                if qnorm > qmax or qnorm == 0:
                    continue
                r = abs(m.StructureFactor(m.Q(h, k, l)))**2
                # test that a peak with non-zero structure factor is indeed
                # allowed
                # Note: the opposite test is not possible because of accidental
                # extinctions
                if not numpy.isclose(r, 0):
                    self.assertIn((h, k, l), hkls, msg=errorinfo)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu

xu.config.VERBOSITY = 0  # make no output during test


class TestMiscutCalc(unittest.TestCase):

    def test_miscut4(self):
        phi = numpy.asarray((0, 45, 90, 135))
        miscut = numpy.random.rand()
        azimuth = numpy.random.rand() * 360
        aom0 = numpy.random.rand() * 45
        aom = miscut * numpy.cos(numpy.radians(phi - azimuth)) + aom0

        om0, p0, mc = xu.analysis.miscut_calc(phi, aom, plot=False)
        self.assertAlmostEqual(om0, aom0, places=5)
        self.assertAlmostEqual(mc, miscut, places=5)
        self.assertAlmostEqual(p0, azimuth, places=5)

    def test_miscut3(self):
        phi = numpy.asarray((0, 60, 120))
        miscut = numpy.random.rand()
        azimuth = numpy.random.rand() * 360
        aom0 = numpy.random.rand() * 45
        aom = miscut * numpy.cos(numpy.radians(phi - azimuth)) + aom0

        om0, p0, mc = xu.analysis.miscut_calc(phi, aom, plot=False)
        self.assertAlmostEqual(om0, aom0, places=5)
        self.assertAlmostEqual(mc, miscut, places=5)
        self.assertAlmostEqual(p0, azimuth, places=5)

    def test_miscut2pom0(self):
        phi = numpy.asarray((0, 90))
        miscut = numpy.random.rand()
        azimuth = numpy.random.rand() * 360
        aom0 = numpy.random.rand() * 45
        aom = miscut * numpy.cos(numpy.radians(phi - azimuth)) + aom0

        om0, p0, mc = xu.analysis.miscut_calc(phi, aom, omega0=aom0,
                                              plot=False)
        self.assertAlmostEqual(om0, aom0, places=5)
        self.assertAlmostEqual(mc, miscut, places=5)
        self.assertAlmostEqual(p0, azimuth, places=5)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu


class TestGridder2DList(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.nx = 10
        # do not change this here unless you fix also the tests cases
        cls.ny = 19
        cls.xmin = 1
        cls.xmax = 10
        cls.x = numpy.linspace(cls.xmin, cls.xmax, num=cls.nx)
        cls.y = cls.x.copy()
        cls.data = numpy.random.rand(cls.nx)
        cls.gridder = xu.Gridder2DList(cls.nx, cls.ny)
        cls.gridder(cls.x, cls.y, cls.data)

    def test_gridder2d_data(self):
        # test shape of data
        self.assertEqual(self.gridder.data.shape[0], self.nx)
        self.assertEqual(self.gridder.data.shape[1], self.ny)
        # test values of data
        aj, ak = numpy.indices((self.nx, self.ny))
        aj, ak = numpy.ravel(aj), numpy.ravel(ak)
        for i in range(self.nx * self.ny):
            j, k = (aj[i], ak[i])
            if k == 2 * j:
                self.assertAlmostEqual(
                    self.gridder.data[j, 2 * j],
                    [self.data[j], ],
                    places=12)
            else:
                self.assertEqual(self.gridder.data[j, k], [])


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2014 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu


class TestNpyGridder1D(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.num = numpy.random.randint(10, 99)
        cls.xmin = 1
        cls.xmax = cls.num
        cls.x = numpy.linspace(cls.xmin, cls.xmax, num=cls.num)
        cls.data = numpy.random.rand(cls.num)
        cls.gridder = xu.npyGridder1D(cls.num)
        cls.gridder(cls.x, cls.data)

    def test_npygridder1d_axis(self):
        hist, bins = numpy.histogram(self.x, bins=self.num)
        x = (bins[0:-1] + bins[1:]) / 2.
        # test length of xaxis
        self.assertEqual(len(self.gridder.xaxis), self.num)
        # test values of xaxis
        for i in range(self.num):
            self.assertAlmostEqual(self.gridder.xaxis[i], x[i], places=12)

    def test_npygridder1d_data(self):
        # test length of data
        self.assertEqual(len(self.gridder.data), self.num)
        # test values of data
        for i in range(self.num):
            self.assertAlmostEqual(
                self.gridder.data[i],
                self.data[i],
                places=12)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015,2018 Dominik Kriegner <dominik.kriegner@gmail.com>

import os.path
import tempfile
import unittest

import numpy
import xrayutilities as xu

xu.config.VERBOSITY = 0  # make no output during test
testfile = 'specmca.spec.gz'
datadir = os.path.join(os.path.dirname(__file__), 'data')
fullfilename = os.path.join(datadir, testfile)


@unittest.skipIf(not os.path.isfile(fullfilename),
                 "additional test data needed (http://xrayutilities.sf.io)")
class TestIO_SPEC(unittest.TestCase):
    dshape = (4001,)
    dmax = 2567926.75
    dmin = 1.0
    motmax = 95.407753
    motmin = 15.40775
    date = 'Mon Nov 04 21:18:05 2013'
    tpos = 2400
    dtpos = 6.0
    scannr = 43
    motorname = 'Nu'
    countername = 'PSDCORR'

    @classmethod
    def setUpClass(cls):
        cls.specfile = xu.io.SPECFile(testfile, path=datadir)
        cls.specfile.Update()  # this should be a noop
        cls.specscan = getattr(cls.specfile, 'scan%d' % cls.scannr)
        cls.specscan.ReadData()
        cls.sdata = cls.specscan.data
        cls.motor, cls.inte = xu.io.getspec_scan(cls.specfile, cls.scannr,
                                                 cls.motorname,
                                                 cls.countername)

    def test_datashape(self):
        self.assertEqual(self.dshape, self.sdata.shape)
        self.assertEqual(self.dshape, self.inte.shape)
        self.assertEqual(self.dshape, self.motor.shape)

    def test_getheader(self):
        self.assertEqual(self.date, self.specscan.getheader_element('D'))

    def test_datavalues(self):
        self.assertAlmostEqual(self.motmax, self.motor.max(), places=6)
        self.assertAlmostEqual(self.motmin, self.motor.min(), places=6)
        self.assertAlmostEqual(self.dmax, self.inte.max(), places=6)
        self.assertAlmostEqual(self.dmin, self.inte.min(), places=6)
        self.assertAlmostEqual(self.dtpos, self.inte[self.tpos], places=6)

    def test_datamethods(self):
        self.assertTrue(numpy.all(self.sdata[self.countername] == self.inte))

    def test_hdf5file(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, 'tmp.h5')
            self.specfile.Save2HDF5(fname)
            h5d = xu.io.geth5_scan(fname, self.scannr)
            self.assertTrue(numpy.all(self.inte == h5d[self.countername]))


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016 Dominik Kriegner <dominik.kriegner@gmail.com>

from multiprocessing import freeze_support

import matplotlib.pyplot as plt
import numpy
import xrayutilities as xu
from mpl_toolkits.axes_grid1 import make_axes_locatable


def main():
    """dummy main function to enable multiprocessing on windows"""
    cryst_size = 40e-9  # meter

    # create Fe BCC (space group nr. 229 Im3m) with a = 2.87 angstrom although
    # this is already predefined as xu.materials.Fe we will repeat here for
    # educational purposes
    FeBCC = xu.materials.Crystal(
        "Fe", xu.materials.SGLattice(229, 2.87,
                                     atoms=[xu.materials.elements.Fe, ],
                                     pos=['2a', ]))

    print("Creating Fe powder ...")
    Fe_powder = xu.simpack.Powder(FeBCC, 1, crystallite_size_gauss=cryst_size)
    pm = xu.simpack.PowderModel(Fe_powder)
    tt = numpy.arange(5, 120, 0.01)
    inte = pm.simulate(tt)

    print(pm)

    # # to create a mixed powder sample one would use
    # Co_powder = xu.simpack.Powder(xu.materials.Co, 5)  # 5 times more Co
    # pmix = xu.simpack.PowderModel(Fe_powder + Co_powder, I0=100)
    # inte = pmix.simulate(tt)
    # pmix.close()  # after end-of-use for cleanup

    ax = pm.plot(tt)
    ax.set_xlim(5, 120)
    pm.close()


if __name__ == '__main__':
    freeze_support()
    main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2013 Dominik Kriegner <dominik.kriegner@gmail.com>
# Copyright (C) 2013 Raphael Grifone <raphael.grifone@esrf.fr>

"""
example to show how to calculate the orientation matrix U from one symmetric
and one asymmetric diffraction peak. The orientation matrix can be used in the
reciprocal space conversion to correct the experimental alignment of the
crystal.

The procedure is as follows:

 1) determine the reciprocal space position without setting a custom U matrix
 2) calculate the orientation matrix
 3) recalculate the reciprocal space positions corrected for a misalignment of
    the lattice from the ideal diffraction condition

In this example only step number 2 is shown. look into other examples to see
how steps 1 and 3 are performed (this depends on the source of data you have)
"""

import numpy
import xrayutilities as xu

# lets assume in step 1) we found the reciprocal space positions of a GaAs
# crystal as

qsym = (0.01822137, -0.013112, 4.45427)  # (004) peak
qasym = (0.0173859, 1.56126512, 3.3450612)  # (113) peak

# the goal is to correct to the diffraction positions in a way that the
# symmetric peak is located on the z-axis and the asymmetric peak is located in
# either the x/z or y/z plane
# we do this by using the coordinate transformations included in xrayutilities

tz = xu.math.AxisToZ(qsym)  # correction of qsym to z-axis
# tz is a coordinate transformation objetct
# calling
tz(qsym)
# should return (~0, ~0, qz)

# now lets correct the sample azimuth in order to have the asymmetric peak in
# the x/z plane, again we make use of coordinate transformations in
# xrayutilities

qasym_tiltcorr = tz(qasym)
phicorr = numpy.arctan2(qasym_tiltcorr[1], qasym_tiltcorr[0])  # in radians
ta = xu.math.ZRotation(-phicorr, deg=False)

# to correctly transform the asymmetric peak to the to x/z plane use
ta(tz(qasym))
# should return (qx, ~0, qz)

# the orientation matrix is then given by

U = numpy.linalg.inv(numpy.dot(ta.matrix, tz.matrix))  # orientation matrix
# the matrix inversion is needed here because of the definition of U,
# which is usually used to calculate the misaligned q from the ideal one

# in case you just want the corrected peak positions you can use again a
# coordinate transformation
t = xu.math.Transform(numpy.linalg.inv(U))

t(qasym)  # will transform qasym to the x/z plane

# for correctly transforming a set of data us U in the call to the
# transformation function for point, linear or area detector
hxrd = xu.HXRD([1, 0, 0], [0, 0, 1])
# qx,qy,qz = hxrd.Ang2Q(om, tt, UB=U)
# or
# qx,qy,qz = hxrd.Ang2Q.linear(om, tt, UB=U)
# or
# qx,qy,qz = hxrd.Ang2Q.area(om, tt, UB=U)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009-2010, 2013
#               Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2009 Mario Keplinger <mario.keplinger@jku.at>
# Copyright (C) 2009-2019 Dominik Kriegner <dominik.kriegner@gmail.com>

import numpy

from . import cxrayutilities, exception, utilities
from .gridder import Gridder, axis, delta, ones


class Gridder2D(Gridder):

    def __init__(self, nx, ny):
        Gridder.__init__(self)

        # check input
        if nx <= 0 or ny <= 0:
            raise exception.InputError('Neither nx nor ny can be smaller'
                                       'than 1!')

        self.xmin = None
        self.ymin = None
        self.xmax = None
        self.ymax = None

        self.nx = nx
        self.ny = ny

        self._allocate_memory()

    def _allocate_memory(self):
        """
        Class method to allocate memory for the gridder based on the nx, ny
        class attributes.
        """

        self._gdata = numpy.zeros((self.nx, self.ny), dtype=numpy.double)
        self._gnorm = numpy.zeros((self.nx, self.ny), dtype=numpy.double)

    def savetxt(self, filename, header=''):
        """
        save gridded data to a txt file with two columns. The first two columns
        are the data coordinates and the last one the corresponding data
        value.

        Parameters
        ----------
        filename :  str
            output filename
        header :    str, optional
            optional header for the data file.
        """
        numpy.savetxt(filename, numpy.vstack((self.xmatrix.flat,
                                              self.ymatrix.flat,
                                              self.data.flat)).T,
                      header=header, fmt='%.6g %.6g %.4g')

    def SetResolution(self, nx, ny):
        """
        Reset the resolution of the gridder. In this case the original data
        stored in the object will be deleted.

        Parameters
        ----------
        nx :	 int
            number of points in x-direction
        ny :	 int
            number of points in y-direction
        """
        self.nx = nx
        self.ny = ny

        self._allocate_memory()

    def __get_xaxis(self):
        return axis(self.xmin, self.xmax, self.nx)

    def __get_yaxis(self):
        return axis(self.ymin, self.ymax, self.ny)

    def __get_xmatrix(self):
        return ones(self.nx, self.ny) * self.xaxis[:, numpy.newaxis]

    def __get_ymatrix(self):
        return ones(self.nx, self.ny) * self.yaxis[numpy.newaxis, :]

    yaxis = property(__get_yaxis)
    xaxis = property(__get_xaxis)
    xmatrix = property(__get_xmatrix)
    ymatrix = property(__get_ymatrix)

    def dataRange(self, xmin, xmax, ymin, ymax, fixed=True):
        """
        define minimum and maximum data range, usually this is deduced
        from the given data automatically, however, for sequential
        gridding it is useful to set this before the first call of the
        gridder. data outside the range are simply ignored

        Parameters
        ----------
        xmin, ymin :    float
            minimum value of the gridding range in x, y
        xmax, ymax :    float
            maximum value of the gridding range in x, y
        fixed :         bool, optional
            flag to turn fixed range gridding on (True (default)) or off
            (False)
        """
        self.fixed_range = fixed
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax

    def _checktransinput(self, x, y, data):
        """
        common checks and reshape commands for the input data. This function
        checks the data type and shape of the input data.
        """
        if not self.keep_data:
            self.Clear()

        x = self._prepare_array(x)
        y = self._prepare_array(y)
        data = self._prepare_array(data)

        if x.size != y.size or y.size != data.size:
            raise exception.InputError("XU.%s: size of given datasets "
                                       "(x, y, data) is not equal!"
                                       % self.__class__.__name__)

        if not self.fixed_range:
            # assume that with setting keep_data the user wants to call the
            # gridder more often and obtain a reasonable result
            self.dataRange(x.min(), x.max(), y.min(), y.max(), self.keep_data)

        return x, y, data

    def __call__(self, x, y, data):
        """
        Perform gridding on a set of data. After running the gridder
        the 'data' object in the class is holding the gridded data.

        Parameters
        ----------
        x :     ndarray
            numpy array of arbitrary shape with x positions
        y :	ndarray
            numpy array of arbitrary shape with y positions
        data :	ndarray
            numpy array of arbitrary shape with data values
        """
        x, y, data = self._checktransinput(x, y, data)
        # remove normalize flag for C-code
        flags = utilities.set_bit(self.flags, 2)
        cxrayutilities.gridder2d(x, y, data, self.nx, self.ny,
                                 self.xmin, self.xmax,
                                 self.ymin, self.ymax,
                                 self._gdata, self._gnorm, flags)


class FuzzyGridder2D(Gridder2D):
    """
    An 2D binning class considering every data point to have a finite area.
    If necessary one data point will be split fractionally over different
    data bins. This is numerically more effort but represents better the
    typical case of a experimental data, which do not represent a mathematical
    point but have a finite size (e.g. X-ray data from a 2D detector or
    reciprocal space maps measured with point/linear detector).

    Currently only a rectangular area can be considered during the gridding.
    """

    def __call__(self, x, y, data, **kwargs):
        """
        Perform gridding on a set of data. After running the gridder
        the 'data' object in the class is holding the gridded data.

        Parameters
        ----------
        x :	ndarray
            numpy array of arbitrary shape with x positions
        y :	ndarray
            numpy array of arbitrary shape with y positions
        data :	ndarray
            numpy array of arbitrary shape with data values
        width :	 float or tuple or list, optional
            width of one data point. If not given half the bin size will be
            used. The width can be given as scalar if it is equal for both data
            dimensions, or as sequence of length 2.
        """

        valid_kwargs = {'width': 'specifiying fuzzy data size'}
        utilities.check_kwargs(kwargs, valid_kwargs,
                               self.__class__.__name__)

        x, y, data = self._checktransinput(x, y, data)

        if 'width' in kwargs:
            try:
                length = len(kwargs['width'])
            except TypeError:
                length = 1
            if length == 2:
                wx, wy = kwargs['width']
            else:
                wx = kwargs['width']
                wy = wx
        else:
            wx = delta(self.xmin, self.xmax, self.nx) / 2.
            wy = delta(self.ymin, self.ymax, self.ny) / 2.
        # remove normalize flag for C-code
        flags = utilities.set_bit(self.flags, 2)
        cxrayutilities.fuzzygridder2d(x, y, data, self.nx, self.ny,
                                      self.xmin, self.xmax,
                                      self.ymin, self.ymax,
                                      self._gdata, self._gnorm, wx, wy, flags)


class Gridder2DList(Gridder2D):

    """
    special version of a 2D gridder which performs no actual averaging of the
    data in one grid/bin but just collects the data-objects belonging to one
    bin for further treatment by the user
    """

    def _allocate_memory(self):
        """
        Class method to allocate memory for the gridder based on the nx, ny
        class attributes.
        """

        self._gdata = numpy.empty((self.nx, self.ny), dtype=list)
        for i in range(self.nx):
            for j in range(self.ny):
                self._gdata[i, j] = []
        self._gnorm = numpy.zeros((self.nx, self.ny), dtype=numpy.int)

    def Clear(self):
        self._allocate_memory()

    def __get_data(self):
        """
        return gridded data, in this special version no normalization is
        defined!
        """
        return self._gdata.copy()

    data = property(__get_data)

    def __call__(self, x, y, data):
        """
        Perform gridding on a set of data. After running the gridder the 'data'
        object in the class is holding the lists of data-objects belonging to
        one bin/grid-point.

        Parameters
        ----------
        x :     ndarray
            numpy array of arbitrary shape with x positions
        y :     ndarray
            numpy array of arbitrary shape with y positions
        data :  ndarray, list or tuple
            data of same length as x, y but of arbitrary type
        """

        x, y, data = self._checktransinput(x, y, data)

        # perform gridding this should be moved to native code if possible
        def gindex(x, min, delt):
            return numpy.round((x - min) / delt).astype(numpy.int)

        xdelta = delta(self.xmin, self.xmax, self.nx)
        ydelta = delta(self.ymin, self.ymax, self.ny)

        for i in range(len(x)):
            xidx = gindex(x[i], self.xmin, xdelta)
            yidx = gindex(y[i], self.ymin, ydelta)
            self._gdata[xidx, yidx].append(data[i])
            self._gnorm[xidx, yidx] += 1



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015 Dominik Kriegner <dominik.kriegner@gmail.com>

import os.path
import unittest

import numpy
import xrayutilities as xu

testfiletmp = 'p08tty_%05d.dat'
testfile = testfiletmp % 29
datadir = os.path.join(os.path.dirname(__file__), 'data')
fullfilename = os.path.join(datadir, testfile)


@unittest.skipIf(not os.path.isfile(fullfilename),
                 "additional test data needed (http://xrayutilities.sf.io)")
class TestIO_TTY(unittest.TestCase):
    dshape = (102,)
    dmax = 1444999.0
    dmin = 314133.0
    motmax = 26.15
    motmin = 13.0
    tpos = 53
    dtpos = 342586.0
    motorname = 'om'
    countername = 'EigerInt'

    @classmethod
    def setUpClass(cls):
        (cls.motor, _), cls.data = xu.io.gettty08_scan(
            testfiletmp, (29, 30), cls.motorname, 'chi', path=datadir)
        cls.inte = cls.data[cls.countername]

    def test_datashape(self):
        self.assertEqual(self.dshape, self.data.shape)
        self.assertEqual(self.dshape, self.motor.shape)

    def test_datavalues(self):
        self.assertAlmostEqual(self.motmax, self.motor.max(), places=6)
        self.assertAlmostEqual(self.motmin, self.motor.min(), places=6)
        self.assertAlmostEqual(self.dmax, self.inte.max(), places=6)
        self.assertAlmostEqual(self.dmin, self.inte.min(), places=6)
        self.assertAlmostEqual(self.dtpos, self.inte[self.tpos], places=6)

    def test_equaldata(self):
        self.assertTrue(numpy.all(self.motor == self.data[self.motorname]))


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2020 Dominik Kriegner <dominik.kriegner@gmail.com>
# Copyright (C) 2020 Mike Moron <mike.moron@tu-dortmund.de>

import numpy as np
import matplotlib.pyplot as plt

import xrayutilities as xu

# the script below currently only works for Amorphous materials.
# file an issue on github if you need/want this for Crystal objects

# create a fictitious LayerStack with thin, rough Layers to illustrate the
# difference between the slicing approach and the usual layered approach
Si = xu.materials.Amorphous('Si', 2285)
SiO2 = xu.materials.Amorphous('SiO2', 1000)
C = xu.materials.Amorphous('C', 800)

s = xu.simpack.Layer(Si, np.inf, roughness=20)
l1 = xu.simpack.Layer(SiO2, 35, roughness=15)
l2 = xu.simpack.Layer(C, 15, roughness=8)

ls = s + l1 + l2

# conventional X-ray reflectivity modelling
m = xu.simpack.SpecularReflectivityModel(ls)
pos, eldens = m.densityprofile(500)

# slice the layerstack into an Amorphous sublayer at every 0.1 angstrom.
# at the top a vacuum layer is added
sls = xu.simpack.effectiveDensitySlicing(ls, 0.1)
ms = xu.simpack.SpecularReflectivityModel(sls)
spos, seldens = ms.densityprofile(500)

# perform simulation and plot simulation and density profile
alpha = np.linspace(0., 5., num=500)

fig = plt.figure()
ax = fig.add_subplot(111)
ax.semilogy(alpha, m.simulate(alpha), label='conventional XRR')
ax.semilogy(alpha, ms.simulate(alpha), label='sliced XRR')
ax.set_xlabel(r'incidence angle')
ax.set_ylabel(r'reflectivity')
ax.legend()

fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(pos, eldens, '.-', label='conventional')
ax.plot(spos, seldens, '.-', label='sliced')  # arbitrary shift for vis.
ax.legend()
ax.set_xlabel(r'z-position')
ax.set_ylabel(r'electron density')

plt.show()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2010 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
xrayutilities derives its own exceptions which are raised
upon wrong input when calling one of xrayutilities functions.
none of the pre-defined exceptions is made for that purpose.
"""

# other used Exception should mainly be the python built-in exceptions
#
# * TypeError
#   Raised when an operation or function is applied to an object of
#   inappropriate type
#
# * ValueError
#   Raised when a operation or function receives an argument that
#   has the right type but an inappropriate value
#
# * UserWarning
#   Base class for warnings generated by user code


class InputError(Exception):

    """
    Exception raised for errors in the input.
    Either wrong datatype not handled by TypeError or missing mandatory
    keyword argument (Note that the obligation to give keyword arguments
    might depend on the value of the arguments itself)

    Parameters
    ----------
    expr :  str
        input expression in which the error occurred
    msg :   str
        explanation of the error
    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2012-2020 Dominik Kriegner <dominik.kriegner@gmail.com>
# Copyright (C) 2012 Tanja Etzelstorfer <tanja.etzelstorfer@jku.at>

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy
import xrayutilities as xu

# plot settings for matplotlib
mpl.rcParams['font.family'] = 'serif'
mpl.rcParams['font.size'] = 20.0
mpl.rcParams['axes.labelsize'] = 'large'

# global setting for the experiment
sample = "rsm"  # sample name used also as file name for the data file

# substrate material used for Bragg peak calculation to correct for
# experimental offsets
Si = xu.materials.Si
Ge = xu.materials.Ge
SiGe = xu.materials.SiGe(1)  # parameter x_Ge = 1

hxrd = xu.HXRD(Si.Q(1, 1, 0), Si.Q(0, 0, 1))

#################################
# Si/SiGe (004) reciprocal space map
omalign = 34.3046  # experimental aligned values
ttalign = 69.1283
# nominal values of the substrate peak
[omnominal, _, _, ttnominal] = hxrd.Q2Ang(Si.Q(0, 0, 4))

# read the data from the xrdml files
om, tt, psd = xu.io.getxrdml_map(sample + '_%d.xrdml.bz2', [1, 2, 3, 4, 5],
                                 path='data')

# determine offset of substrate peak from experimental values (optional)
omalign, ttalign, p, cov = xu.analysis.fit_bragg_peak(
    om, tt, psd, omalign, ttalign, hxrd, plot=False)

# convert angular coordinates to reciprocal space + correct for offsets
qx, qy, qz = hxrd.Ang2Q(om, tt, delta=[omalign - omnominal,
                                       ttalign - ttnominal])

# calculate data on a regular grid of 200x201 points
gridder = xu.Gridder2D(200, 600)
gridder(qy, qz, psd)
INT = xu.maplog(gridder.data.transpose(), 6, 0)

# plot the intensity as contour plot
plt.figure()
cf = plt.contourf(gridder.xaxis, gridder.yaxis, INT, 100, extend='min')
plt.xlabel(r'$Q_{[110]}$ ($\mathrm{\AA}^{-1}$)')
plt.ylabel(r'$Q_{[001]}$ ($\mathrm{\AA}^{-1}$)')
cb = plt.colorbar(cf)
cb.set_label(r"$\log($Int$)$ (cps)")

tr = SiGe.RelaxationTriangle([0, 0, 4], Si, hxrd)
# plt.plot(tr[0], tr[1], 'ok')
plt.tight_layout()

# line cut with integration along 2theta to remove beam footprint broadening
qycpos, qzcpos = [0, 4.5]
omp, _, _, ttp = hxrd.Q2Ang(0, qycpos, qzcpos, trans=False)
qzc, qzint, cmask = xu.analysis.get_radial_scan(
    [qy, qz], psd, (qycpos, qzcpos), 1001, 0.155, intdir='2theta')

# show used data on the reciprocal space map
plt.tricontour(qy, qz, cmask, (0.999,), colors='r')

# show possible integration directions
f = 1.02
plt.plot([qycpos/f, qycpos*f], [qzcpos/f, qzcpos*f], label='radial')
a = numpy.radians(numpy.linspace(- 2, 2, 100))
plt.plot(qycpos*numpy.cos(a)+qzcpos*numpy.sin(a),
         -qycpos*numpy.sin(a)+qzcpos*numpy.cos(a), label='Omega')
qx, qy, qz = hxrd.Ang2Q(omp, ttp+numpy.linspace(-2, 2, 100))
plt.plot(qy, qz, label='2Theta')
plt.legend()

# plot line cut
plt.figure()
plt.semilogy(qzc, qzint)
plt.xlabel(r'scattering angle (deg)')
plt.ylabel(r'intensity (arb. u.)')
plt.tight_layout()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2019 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu

try:
    import lmfit
except ImportError:
    lmfit = None


@unittest.skipIf(lmfit is None, "the lmfit Python package is needed")
class Test_SpecularReflectivityModel(unittest.TestCase):
    # define used layer stack
    lSiO2 = xu.simpack.Layer(xu.materials.SiO2, numpy.inf, roughness=3.0)
    lRu = xu.simpack.Layer(xu.materials.Ru, 50, roughness=2.5)
    rho_cf = 0.5*8900 + 0.5*7874
    mat_cf = xu.materials.Amorphous('CoFe', rho_cf)
    lCoFe = xu.simpack.Layer(mat_cf, 30, roughness=5.2)

    # simulation parameters
    kwargs = dict(I0=1e6, background=2, sample_width=10, beam_width=0.2,
                  energy='MoKa1')

    @classmethod
    def setUpClass(cls):
        cls.model = xu.simpack.SpecularReflectivityModel(cls.lSiO2, cls.lRu,
                                                         cls.lCoFe,
                                                         **cls.kwargs)
        cls.fm = xu.simpack.FitModel(cls.model)
        cls.ai = numpy.arange(0.1, 3, 0.005)

    def test_Calculation(self):
        sim = self.model.simulate(self.ai)
        self.assertEqual(len(sim), len(self.ai))
        self.assertTrue(numpy.all(sim >= self.kwargs['background']))
        self.assertTrue(numpy.all(sim <= self.kwargs['I0']))

    def test_FitModel_eval(self):
        sim1 = self.model.simulate(self.ai)
        p = self.fm.make_params()
        sim2 = self.fm.eval(p, x=self.ai)
        for v1, v2 in zip(sim1, sim2):
            self.assertAlmostEqual(v1, v2, places=10)

    def test_densityprofile(self):
        N = 123
        z, d = self.model.densityprofile(N)
        self.assertEqual(len(z), N)
        self.assertEqual(len(d), N)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2012 Dominik Kriegner <dominik.kriegner@gmail.com>

import numpy
import xrayutilities as xu

# config for kappa geometry
# set it either in the script or include it in on of the config files

xu.config.KAPPA_PLANE = 'zy'
xu.config.KAPPA_ANGLE = -60

# kappa goniometer as shown in
# http://en.wikipedia.org/wiki/File:Kappa_goniometer_animation.ogg
qconv = xu.experiment.QConversion(['z+', 'k+', 'z+'], ['z+'], (1, 0, 0))

print(qconv)

print("angles: 0, 0, 0, 90")
(qx, qy, qz) = qconv(0, 0, 0, 90)
print("Q= %6.3f %6.3f %6.3f (Abs: %6.3f)"
      % (qx, qy, qz, numpy.linalg.norm((qx, qy, qz))))

print("angles: 90, 0, 0, 90")
(qx, qy, qz) = qconv(90, 0, 0, 90)
print("Q= %6.3f %6.3f %6.3f (Abs: %6.3f)"
      % (qx, qy, qz, numpy.linalg.norm((qx, qy, qz))))

print("angles: 0, 90, 0, 90")
(qx, qy, qz) = qconv(0, 90, 0, 90)
print("Q= %6.3f %6.3f %6.3f (Abs: %6.3f)"
      % (qx, qy, qz, numpy.linalg.norm((qx, qy, qz))))



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2012 Dominik Kriegner <dominik.kriegner@gmail.com>

import os

import numpy
import xrayutilities as xu

# import matplotlib.pyplot as plt

# create material
Calcite = xu.materials.Crystal.fromCIF(os.path.join("data", "Calcite.cif"))

# experiment class with some weird directions
expcal = xu.HXRD(Calcite.Q(-2, 1, 9), Calcite.Q(1, -1, 4))

powder_cal = xu.simpack.PowderDiffraction(Calcite)
print(powder_cal)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2012 Dominik Kriegner <dominik.kriegner@gmail.com>

import numpy
import xrayutilities as xu

# defining material and experimental setup
InAs = xu.materials.InAs
energy = 8048  # eV

# calculate the structure factor for InAs (111) (222) (333)
hkllist = [[1, 1, 1], [2, 2, 2], [3, 3, 3]]
for hkl in hkllist:
    qvec = InAs.Q(hkl)
    F = InAs.StructureFactor(qvec, energy)
    print(" |F| = %8.3f" % numpy.abs(F))



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2013,2014 Dominik Kriegner <dominik.kriegner@gmail.com>


"""
This example shows the use of the Ang2HKL function
used to convert angles to HKL coordinates of a certain material

The orientation of the crystal is specified by the two directions given to the
HXRD class.

The same example is shown twice using different goniometer definitions with
different incidence beam directions. The different definition changes the
momentum transfer Q (only its orientation), however both yields the same HKL
upon back and force conversion to angular and HKL-space.
"""

import xrayutilities as xu

# material used in this example
mat = xu.materials.Si

H = 3
K = 3
L = 1

# define experiment geometry
hxrd = xu.HXRD(mat.Q(1, 1, -2), mat.Q(1, 1, 1))

# calculate angles
[om, chi, phi, tt] = hxrd.Q2Ang(mat.Q(H, K, L))

# perform conversion to reciprocal space
print(hxrd.Transform(mat.Q(H, K, L)))
print(hxrd.Ang2Q(om, tt))

# perform conversion to HKL coordinates
if chi != 0 or phi != 0:
    print("Geometry not enough for full description -> HKL will be wrong")
print(hxrd.Ang2HKL(om, tt, mat=mat))

print("--------------")
# example with custom qconv
qconv = xu.experiment.QConversion(['z+', 'x+', 'y-'], 'z+', [1, 0, 0])

# material used in this example
mat = xu.materials.Si

# define experiment geometry
hxrd = xu.HXRD(mat.Q(1, 1, -2), mat.Q(1, 1, 1), qconv=qconv)

# calculate angles
# IMPORTANT: the Q2Ang function is not influenced by the qconv argument. A
# proper calculation of the angles needs the Q2AngFit function described in
# the example file 'xrayutilities_q2ang_general.py'
[om, chi, phi, tt] = hxrd.Q2Ang(mat.Q(H, K, L))

# perform conversion to reciprocal space
print(hxrd.Transform(mat.Q(H, K, L)))
print(hxrd.Ang2Q(om, chi, phi, tt))

# perform conversion ro HKL coordinates
print(hxrd.Ang2HKL(om, chi, phi, tt, mat=mat))



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009 Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2010-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
module with vector operations for vectors of size 3,
since for so short vectors numpy does not give the best performance explicit
implementation of the equations is performed together with error checking to
ensure vectors of length 3.
"""

import math
import re

import numpy

from ..exception import InputError

circleSyntax = re.compile("[xyz][+-]")


def _checkvec(v):
    if isinstance(v, (list, tuple, numpy.ndarray)):
        vtmp = numpy.asarray(v, dtype=numpy.double)
    else:
        raise TypeError("Vector must be a list, tuple or numpy array")
    return vtmp


def VecNorm(v):
    """
    Calculate the norm of a vector.

    Parameters
    ----------
    v :     list or array-like
        input vector(s), either one vector or an array of vectors with shape
        (n, 3)

    Returns
    -------
    float or ndarray
        vector norm, either a single float or shape (n, )
    """
    if isinstance(v, numpy.ndarray):
        if len(v.shape) >= 2:
            return numpy.linalg.norm(v, axis=-1)
    if len(v) != 3:
        raise ValueError("Vector must be of length 3, but has length %d!"
                         % len(v))
    return math.sqrt(v[0]**2 + v[1]**2 + v[2]**2)


def VecUnit(v):
    """
    Calculate the unit vector of v.

    Parameters
    ----------
    v :     list or array-like
        input vector(s), either one vector or an array of vectors with shape
        (n, 3)

    Returns
    -------
    ndarray
        unit vector of `v`, either shape (3, ) or (n, 3)
    """
    vtmp = _checkvec(v)
    if len(vtmp.shape) == 1:
        return vtmp / VecNorm(vtmp)
    else:
        return vtmp / VecNorm(vtmp)[..., numpy.newaxis]


def VecDot(v1, v2):
    """
    Calculate the vector dot product.

    Parameters
    ----------
    v1, v2 :    list or array-like
        input vector(s), either one vector or an array of vectors with shape
        (n, 3)

    Returns
    -------
    float or ndarray
        innter product of the vectors, either a single float or (n, )
    """
    if isinstance(v1, numpy.ndarray):
        if len(v1.shape) >= 2:
            return numpy.einsum('...i, ...i', v1, v2)
    if len(v1) != 3 or len(v2) != 3:
        raise ValueError("Vectors must be of size 3! (len(v1)=%d len(v2)=%d)"
                         % (len(v1), len(v2)))

    return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]


def VecCross(v1, v2, out=None):
    """
    Calculate the vector cross product.

    Parameters
    ----------
    v1, v2 :    list or array-like
        input vector(s), either one vector or an array of vectors with shape
        (n, 3)
    out :       list or array-like, optional
        output vector

    Returns
    -------
    ndarray
        cross product either of shape (3, ) or (n, 3)
    """
    if isinstance(v1, numpy.ndarray):
        if len(v1.shape) >= 2 or len(v2.shape) >= 2:
            return numpy.cross(v1, v2)
    if len(v1) != 3 or len(v2) != 3:
        raise ValueError("Vectors must be of size 3! (len(v1)=%d len(v2)=%d)"
                         % (len(v1), len(v2)))
    if out is None:
        out = numpy.empty(3)
    out[0] = v1[1] * v2[2] - v1[2] * v2[1]
    out[1] = v1[2] * v2[0] - v1[0] * v2[2]
    out[2] = v1[0] * v2[1] - v1[1] * v2[0]
    return out


def VecAngle(v1, v2, deg=False):
    """
    calculate the angle between two vectors. The following
    formula is used
    v1.v2 = norm(v1)*norm(v2)*cos(alpha)

    alpha = acos((v1.v2)/(norm(v1)*norm(v2)))

    Parameters
    ----------
    v1, v2 :    list or array-like
        input vector(s), either one vector or an array of vectors with shape
        (n, 3)
    deg:        bool, optional
        True: return result in degree, False: in radiants (default: False)

    Returns
    -------
    float or ndarray
        the angle included by the two vectors `v1` and `v2`, either a single
        float or an array with shape (n, )
    """
    u1 = VecNorm(v1)
    u2 = VecNorm(v2)

    if isinstance(u1, numpy.ndarray) or isinstance(u2, numpy.ndarray):
        s = VecDot(v1, v2) / u1 / u2
        s[numpy.abs(s) > 1.0] = numpy.sign(s[numpy.abs(s) > 1.0]) * 1.0
        alpha = numpy.arccos(s)
        if deg:
            alpha = numpy.degrees(alpha)
    else:
        alpha = math.acos(max(min(1., VecDot(v1, v2) / u1 / u2), -1.))
        if deg:
            alpha = math.degrees(alpha)

    return alpha


def distance(x, y, z, point, vec):
    """
    calculate the distance between the point (x, y, z) and the line defined by
    the point and vector vec

    Parameters
    ----------
    x :     float or ndarray
        x coordinate(s) of the point(s)
    y :     float or ndarray
        y coordinate(s) of the point(s)
    z :     float or ndarray
        z coordinate(s) of the point(s)
    point : tuple, list or ndarray
        3D point on the line to which the distance should be calculated
    vec :   tuple, list or ndarray
        3D vector defining the propergation direction of the line
    """
    coords = numpy.vstack((numpy.ravel(x) - point[0],
                           numpy.ravel(y) - point[1],
                           numpy.ravel(z) - point[2])).T
    ret = VecNorm(VecCross(coords, numpy.asarray(vec)))/VecNorm(vec)
    if isinstance(x, numpy.ndarray):
        ret = ret.reshape(x.shape)
    else:
        ret = ret[0]
    return ret


def getVector(string):
    """
    returns unit vector along a rotation axis given in the syntax
    'x+' 'z-' or equivalents

    Parameters
    ----------
    string:     str
        vector string following the synthax [xyz][+-]

    Returns
    -------
    ndarray
        vector along the given direction
    """

    if len(string) != 2:
        raise InputError("wrong length of string for conversion given")
    if not circleSyntax.search(string):
        raise InputError("getVector: incorrect string syntax (%s)" % string)

    if string[0] == 'x':
        v = [1., 0, 0]
    elif string[0] == 'y':
        v = [0, 1., 0]
    elif string[0] == 'z':
        v = [0, 0, 1.]
    else:
        raise InputError("wrong first character of string given "
                         "(needs to be one of x, y, z)")

    if string[1] == '+':
        v = numpy.asarray(v) * (+1)
    elif string[1] == '-':
        v = numpy.asarray(v) * (-1)
    else:
        raise InputError("wrong second character of string given "
                         "(needs to be + or -)")

    return v


def getSyntax(vec):
    """
    returns vector direction in the syntax
    'x+' 'z-' or equivalents
    therefore works only for principle vectors of the coordinate system
    like e.g. [1, 0, 0] or [0, 2, 0]

    Parameters
    ----------
    vec :   list or array-like
        vector of length 3

    Returns
    -------
    str
        vector string following the synthax [xyz][+-]
    """
    v = _checkvec(vec)
    if len(v) != 3:
        raise InputError("no valid 3D vector given")

    x = [1, 0, 0]
    y = [0, 1, 0]
    z = [0, 0, 1]

    if numpy.isclose(VecNorm(numpy.cross(numpy.cross(x, y), v)), 0):
        if v[2] >= 0:
            string = 'z+'
        else:
            string = 'z-'
    elif numpy.isclose(VecNorm(numpy.cross(numpy.cross(x, z), v)), 0):
        if v[1] >= 0:
            string = 'y+'
        else:
            string = 'y-'
    elif numpy.isclose(VecNorm(numpy.cross(numpy.cross(y, z), v)), 0):
        if v[0] >= 0:
            string = 'x+'
        else:
            string = 'x-'
    else:
        raise InputError("no valid 3D vector given")

    return string



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016 Dominik Kriegner <dominik.kriegner@gmail.com>

import time

import xrayutilities as xu
from matplotlib.pylab import *

mpl.rcParams['font.size'] = 16.0


en = 'CuKa1'  # eV
lam = xu.en2lam(en)
resol = 2*pi/4998  # resolution in q; to suppress buffer oscillations
H, K, L = (0, 0, 4)
qz = linspace(4.0, 5.0, 3000)

sub = xu.simpack.Layer(xu.materials.Si, inf)
#                                                   xfrom xto nsteps thickness
buf1 = xu.simpack.GradedLayerStack(xu.materials.SiGe, 0.2, 0.8, 100, 10000,
                                   relaxation=1.0)
buf2 = xu.simpack.Layer(xu.materials.SiGe(0.8), 4997.02, relaxation=1.0)
lay1 = xu.simpack.Layer(xu.materials.SiGe(0.6), 49.73, relaxation=0.0)
lay2 = xu.simpack.Layer(xu.materials.SiGe(1.0), 45.57, relaxation=0.0)
# create superlattice stack
# lattice param. are adjusted to the relaxation parameter of the layers
# Note that to create a superlattice you can use summation and multiplication
pls = xu.simpack.PseudomorphicStack001('SL 5/5', sub+buf1+buf2+5*(lay1+lay2))

# calculate incidence angle for dynamical diffraction models
qx = sqrt(sub.material.Q(H, K, L)[0]**2 + sub.material.Q(H, K, L)[1]**2)
ai = xu.simpack.coplanar_alphai(qx, qz, en)
resolai = abs(xu.simpack.coplanar_alphai(qx, mean(qz) + resol, en) -
              xu.simpack.coplanar_alphai(qx, mean(qz), en))

# comparison of different diffraction models
# simplest kinematical diffraction model
t0 = time.time()
mk = xu.simpack.KinematicalModel(pls, energy=en, resolution_width=resol)
Ikin = mk.simulate(qz, hkl=(0, 0, 4))
t1 = time.time()
print("%.3f sec for kinematical calculation" % (t1-t0))

# kinematical multibeam model
t0 = time.time()
mk = xu.simpack.KinematicalMultiBeamModel(pls, energy=en,
                                          surface_hkl=(0, 0, 1),
                                          resolution_width=resol)
Imult = mk.simulate(qz, hkl=(H, K, L), refraction=True)
t1 = time.time()
print("%.3f sec for kinematical multibeam calculation" % (t1-t0))

# general 2-beam theory based dynamical diffraction model
t0 = time.time()
qGe220 = linalg.norm(xu.materials.Ge.Q(2, 2, 0))
thMono = arcsin(qGe220 * lam / (4*pi))
md = xu.simpack.DynamicalModel(pls, energy=en, resolution_width=resolai,
                               Cmono=cos(2 * thMono),
                               polarization='both')
Idyn = md.simulate(ai, hkl=(H, K, L))
t1 = time.time()
print("%.3f sec for acurate dynamical calculation" % (t1-t0))

# plot of calculated intensities
figure('XU-simpack SiGe SL')
clf()
semilogy(qz, Ikin, label='kinematical')
semilogy(qz, Imult, label='multibeam')
semilogy(xu.simpack.get_qz(qx, ai, en), Idyn, label='full dynamical')
vlines([4*2*pi/l.material.a3[-1] for l in pls[-2:]], 1e-9, 1,
       linestyles='dashed', label="kin. peak-pos")
legend(fontsize='small')
xlim(qz.min(), qz.max())
xlabel(r'Qz ($1/\mathrm{\AA}$)')
ylabel('Intensity (arb. u.)')
tight_layout()
show()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2019 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu


class TestMathVector(unittest.TestCase):
    def test_distance(self):
        vec = (3, 1, 1)
        point = (-4, -5, -1)
        x, y, z = (-6, 1, 21)
        d = xu.math.vector.distance(numpy.array((x, x)),
                                    numpy.array((y, y)),
                                    numpy.array((z, z)),
                                    point, vec)
        self.assertEqual(d.size, 2)
        self.assertAlmostEqual(d[0], 4*numpy.sqrt(30), places=10)
        self.assertAlmostEqual(d[1], 4*numpy.sqrt(30), places=10)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2014 Dominik Kriegner <dominik.kriegner@gmail.com>

import os

import matplotlib.pyplot as plt
import numpy
import xrayutilities as xu

# load powder diffraction cif data file
# here use e.g. cif-file from ICDD
pd = xu.io.pdCIF(os.path.join('data', '04-003-0996.cif'))
# plot data
plt.figure()
plt.semilogy(
    pd.data['_pd_meas_2theta_scan'],
    pd.data['_pd_meas_intensity_total'])
plt.xlim(0, 90)
plt.ylim(1e-1, 1.5e3)
plt.ylabel('Intensity Sb$_2$Te$_3$ (arb.u.)')
plt.xlabel('scattering angle (deg)')

# load materials
# structure cif from Pearson's crystal data database
st = xu.materials.Crystal.fromCIF(os.path.join('data', '1216385.cif'))
pst = xu.simpack.PowderDiffraction(st, tt_cutoff=90)

height = 0.05
# peak positions of Sb2Te3
mask = pst.data / pst.data.max() > 0.001
plt.bar(pst.ang[mask] * 2, numpy.ones(mask.sum()) * height, width=0.4,
        bottom=0.13, linewidth=0, color='r', align='center',
        orientation='vertical')



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
Module provides functions to convert a q-vector from reciprocal space to
angular space. a simple implementation uses scipy optimize routines to perform
a fit for a arbitrary goniometer.

The user is, however, expected to use the bounds variable to put restrictions
to the number of free angles to obtain reproducible results. In general only 3
angles are needed to fit an arbitrary q-vector (2 sample + 1 detector angles or
1 sample + 2 detector). More complicated restrictions can be implemented using
the lmfit package. (done upon request!)

The function is based on a fitting routine. For a specific goniometer also
analytic expressions from literature can be used as they are implemented in the
predefined experimental classes HXRD, NonCOP, and GID.
"""

import numbers

import numpy
import scipy.optimize

from . import config, math
from .exception import InputError


def _makebounds(boundsin):
    """
    generate proper bounds for scipy.optimize.minimize function
    from a list/tuple of more convenient bounds.

    Parameters
    ----------
    boundsin :  list or tuple or array-like
        bounds, or fixed values. the number of entries needs to be equal to the
        number of angle in the goniometer given to the q2ang_general function
        example input for four gonimeter angles: ((0, 90), 0, (0, 180), (0,
        90))

    Returns
    -------
    scipy.optimize.Bounds object
        bounds to be handed over to the scipy.minimize routine. The function
        will expand fixed values to two equal bounds
    constraints, list
        list of equivility constraints for fixed values
    """
    lb = []
    ub = []
    constraints = []
    for j, b in enumerate(boundsin):
        if isinstance(b, (tuple, list, numpy.ndarray)):
            if len(b) == 2:
                lb.append(b[0])
                ub.append(b[1])
            elif len(b) == 1:
                # due to a bug in scipy >= 1.5.0 we need to allow a small
                # variation
                lb.append(b[0] - config.EPSILON)
                ub.append(b[0] + config.EPSILON)
                # to really fix the parameter we create an equivalent
                # constraint; see scipy/scipy#12433
                constraints.append(dict(type='eq',
                                        fun=lambda x, j=j, v=b[0]: x[j] - v,
                                        # lambda j=j to bind variable by value
                                        ))
            else:
                raise InputError('bound values must have two or one elements')
        elif isinstance(b, numbers.Number):
            # due to a bug in scipy >= 1.5.0 we need to allow a small variation
            lb.append(b - config.EPSILON)
            ub.append(b + config.EPSILON)
            # to really fix the parameter we create an equivalent
            # constraint; see scipy/scipy#12433
            constraints.append(dict(type='eq',
                                    fun=lambda x, j=j, v=b: x[j] - v,
                                    # lambda j=j to bind variable by value
                                    ))
        elif b is None:
            lb.append(-np.inf)
            ub.append(np.inf)
        else:
            raise InputError('bound value is of invalid type (%s)' % type(b))

    return scipy.optimize.Bounds(lb, ub), constraints


def _errornorm_q2ang(angles, qvec, hxrd, U=numpy.identity(3)):
    """
    function to determine the offset in the qposition calculated from
    a set of experimental angles and the given vector

    Parameters
    ----------
    angles :    iterable
        iterable object with angles of the goniometer
    qvec :      list or tuple or array-like
        vector with three q-coordinates
    hxrd :      Experiment
        experiment class to be used for the q calculation
    U :         array-like, optional
        orientation matrix

    Returns
    -------
    error : float
        q-space error between the current fit-guess and the user-specified
        position
    """

    qcalc = hxrd.Ang2Q.point(*angles, UB=U)
    dq = numpy.linalg.norm(qcalc - qvec)
    return dq


def incidenceAngleConst(angles, alphai, xrd):
    """
    helper function for an pseudo-angle constraint of the incidence angle. Can
    be used together with the Q2AngFit-routine in the 'constraints' argument.
    An example use case scenario to fix the incidence angle to 1 degree would
    be:
    constraints={'type': 'eq', 'fun': lambda a: incidenceAngleConst(a, 1, xrd)}

    Parameters
    ----------
    angles :    iterable
        fit parameters of Q2AngFit
    alphai :    float
        the incidence angle which should be fixed
    xrd :       Experiment
        the Experiment object to use for qconversion
    """
    qconv = xrd._A2QConversion
    ndirlab = qconv.transformSample2Lab(xrd.Transform(xrd.ndir), *angles)
    ai = 90 - math.VecAngle(-qconv.r_i, ndirlab, deg=True) - alphai
    return ai


def exitAngleConst(angles, alphaf, xrd):
    """
    helper function for an pseudo-angle constraint of the exit angle. Can be
    used together with the Q2AngFit-routine in the 'constraints' argument. An
    example use case scenario to fix the exit angle to 1 degree would be:
    constraints={'type': 'eq', 'fun': lambda a: exitAngleConst(a, 1, xrd)}

    Parameters
    ----------
    angles :    iterable
        fit parameters of Q2AngFit
    alphaf :    float
        the exit angle which should be fixed
    xrd :       Experiment
        the Experiment object to use for qconversion
    """
    qconv = xrd._A2QConversion
    # calc kf
    detangles = [a for a in angles[-len(qconv.detectorAxis):]]
    kf = qconv.getDetectorPos(*detangles)
    if numpy.linalg.norm(kf) == 0:
        af = 0
    else:
        ndirlab = qconv.transformSample2Lab(xrd.Transform(xrd.ndir), *angles)
        af = 90 - math.VecAngle(kf, ndirlab, deg=True) - alphaf
    return af


def Q2AngFit(qvec, expclass, bounds=None, ormat=numpy.identity(3),
             startvalues=None, constraints=[]):
    """
    Functions to convert a q-vector from reciprocal space to angular space.
    This implementation uses scipy optimize routines to perform a fit for a
    goniometer with arbitrary number of goniometer angles.

    The user *must* use the bounds variable to put
    restrictions to the number of free angles to obtain reproducible results.
    In general only 3 angles are needed to fit an arbitrary q-vector (2 sample
    + 1 detector angles or 1 sample + 2 detector).

    Parameters
    ----------
    qvec :      tuple or list or array-like
        q-vector for which the angular positions should be calculated
    expclass :  Experiment
        experimental class used to define the goniometer for which the angles
        should be calculated.
    bounds :    tuple or list
        bounds of the goniometer angles. The number of bounds must correspond
        to the number of goniometer angles in the expclass.  Angles can also be
        fixed by supplying only one value for a particular angle. e.g.: ((low,
        up), fix, (low2, up2), (low3, up3))
    ormat :     array-like
        orientation matrix of the sample to be used in the conversion
    startvalues :   array-like
        start values for the fit, which can significantly speed up the
        conversion. The number of values must correspond to the number of
        angles in the goniometer of the expclass
    constraints :   list
        sequence of constraint dictionaries. This allows applying arbitrary
        (e.g. pseudo-angle) contraints by supplying according constraint
        functions. An entry of the constraints argument must be a dictionary
        with at least the 'type' and 'fun' set. 'type' can be either 'eq' or
        'ineq' for equality or inequality constraints. 'fun' must be a callable
        function which for 'eq'-constraints returns 0 when the equality
        condition is fulfilled (see constraints documentation in
        scipy.optimize.minimize for details). The supplied function will be
        called with the arguments gonimeter angle list as argument. Typically
        this means you will have to use a lambda function.

    Returns
    -------
    fittedangles :  list
        list of fitted goniometer angles
    qerror :        float
        error in reciprocal space
    errcode :       int
        error-code of the scipy minimize function. for a successful fit the
        error code should be <=2
    """

    # check input parameters
    if len(qvec) != 3:
        raise ValueError("XU.Q2AngFit: length of given q-vector is not 3 "
                         "-> invalid")
    lqvec = numpy.asarray(qvec)

    qconv = expclass._A2QConversion
    nangles = len(qconv.sampleAxis) + len(qconv.detectorAxis)

    # generate starting position for optimization
    if startvalues is None:
        start = numpy.zeros(nangles)
    else:
        start = startvalues

    # check bounds
    if bounds is None:
        bounds = numpy.zeros(2 * nangles) - 180.
        bounds[::2] = 180.
        bounds.shape = (nangles, 2)
    elif len(bounds) != nangles:
        raise ValueError("XU.Q2AngFit: number of specified bounds invalid")

    sbounds, boundconstraints = _makebounds(bounds)
    sconstraints = list(constraints) + boundconstraints
    # perform optimization
    res = scipy.optimize.minimize(_errornorm_q2ang, start,
                                  args=(lqvec, expclass, ormat),
                                  method='SLSQP', bounds=sbounds,
                                  constraints=sconstraints,
                                  options={'maxiter': 1000,
                                           'eps': config.EPSILON,
                                           'ftol': config.EPSILON})

    x, errcode, qerror = (res.x, res.status, res.fun)
    if qerror >= 1e-7:
        if config.VERBOSITY >= config.DEBUG:
            print("XU.Q2AngFit: info: need second run")
        # make a second run
        res = scipy.optimize.minimize(_errornorm_q2ang, res.x,
                                      args=(lqvec, expclass, ormat),
                                      method='SLSQP',
                                      bounds=sbounds,
                                      constraints=sconstraints,
                                      options={'maxiter': 1000,
                                               'eps': config.EPSILON,
                                               'ftol': config.EPSILON})
        x, errcode, qerror = (res.x, res.status, res.fun)

    if ((config.VERBOSITY >= config.DEBUG) or (qerror > 10*config.EPSILON and
                                               config.VERBOSITY >=
                                               config.INFO_LOW)):
        print("XU.Q2AngFit: q-error=%.4g with error-code %d (%s)"
              % (qerror, errcode, res.message))

    return x, qerror, errcode



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2014-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

import copy
import unittest

import numpy
import xrayutilities as xu
from scipy.integrate import nquad, quad


class TestMathFunctions(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        amp = numpy.random.rand() + 0.1
        fwhm = numpy.random.rand() * 1.5 + 0.1
        cls.x = numpy.arange(-3, 3, 0.0003)
        cls.p = [0., fwhm, amp, 0.]
        cls.p2d = [0., 0., fwhm, fwhm, amp, 0.,
                   2 * numpy.pi * numpy.random.rand()]
        cls.sigma = fwhm / (2 * numpy.sqrt(2 * numpy.log(2)))

    def test_gauss1dwidth(self):
        p = numpy.copy(self.p)
        p[1] = self.sigma
        f = xu.math.Gauss1d(self.x, *p)
        fwhm = xu.math.fwhm_exp(self.x, f)
        self.assertAlmostEqual(fwhm, self.p[1], places=4)

    def test_lorentz1dwidth(self):
        p = numpy.copy(self.p)
        f = xu.math.Lorentz1d(self.x, *p)
        fwhm = xu.math.fwhm_exp(self.x, f)
        self.assertAlmostEqual(fwhm, self.p[1], places=4)

    def test_pvoigt1dwidth(self):
        p = list(numpy.copy(self.p))
        p += [numpy.random.rand(), ]
        f = xu.math.PseudoVoigt1d(self.x, *p)
        fwhm = xu.math.fwhm_exp(self.x, f)
        self.assertAlmostEqual(fwhm, self.p[1], places=4)

    def test_normedgauss1d(self):
        p = numpy.copy(self.p)
        p[1] = self.sigma
        f = xu.math.Gauss1d(self.x, *p) / xu.math.Gauss1dArea(*p)
        norm = xu.math.NormGauss1d(self.x, p[0], p[1])
        self.assertAlmostEqual(numpy.sum(numpy.abs(f - norm)), 0, places=6)

    def test_gauss1darea(self):
        p = numpy.copy(self.p)
        p[1] = self.sigma
        area = xu.math.Gauss1dArea(*p)
        (numarea, err) = quad(
            xu.math.Gauss1d, -numpy.inf, numpy.inf, args=tuple(p))
        digits = int(numpy.abs(numpy.log10(err))) - 3
        self.assertTrue(digits >= 3)
        self.assertAlmostEqual(area, numarea, places=digits)

    def test_lorentz1darea(self):
        p = numpy.copy(self.p)
        area = xu.math.Lorentz1dArea(*p)
        (numarea, err) = quad(
            xu.math.Lorentz1d, -numpy.inf, numpy.inf, args=tuple(p))
        digits = int(numpy.abs(numpy.log10(err))) - 3
        self.assertTrue(digits >= 3)
        self.assertAlmostEqual(area, numarea, places=digits)

    def test_pvoigt1darea(self):
        p = list(numpy.copy(self.p))
        p += [numpy.random.rand(), ]
        area = xu.math.PseudoVoigt1dArea(*p)
        (numarea, err) = quad(
            xu.math.PseudoVoigt1d, -numpy.inf, numpy.inf, args=tuple(p))
        digits = int(numpy.abs(numpy.log10(err))) - 3
        self.assertTrue(digits >= 3)
        self.assertAlmostEqual(area, numarea, places=digits)

    def test_gauss2darea(self):
        p = numpy.copy(self.p2d)
        p[2] = self.sigma
        p[3] = (numpy.random.rand() + 0.1) * self.sigma
        area = xu.math.Gauss2dArea(*p)
        (numarea, err) = nquad(xu.math.Gauss2d, [[-numpy.inf, numpy.inf],
                                                 [-numpy.inf, numpy.inf]],
                               args=tuple(p))
        digits = int(numpy.abs(numpy.log10(err))) - 3
        self.assertTrue(digits >= 3)
        self.assertAlmostEqual(area, numarea, places=digits)

    def test_derivatives(self):
        eps = 1e-9
        # generate test input parameters (valid for Gauss, Lorentz, and Voigt)
        p = list(numpy.copy(self.p))
        p[1] = self.sigma
        p[3] = numpy.random.rand()
        p.append(numpy.random.rand())
        params = [self.x, ] + p

        functions = [(xu.math.Gauss1d, xu.math.Gauss1d_der_x,
                      xu.math.Gauss1d_der_p),
                     (xu.math.Lorentz1d, xu.math.Lorentz1d_der_x,
                      xu.math.Lorentz1d_der_p),
                     (xu.math.PseudoVoigt1d, xu.math.PseudoVoigt1d_der_x,
                      xu.math.PseudoVoigt1d_der_p)]

        # test all derivates by benchmarking against a simple finite difference
        # calculation
        for f, fdx, fdp in functions:
            deriv = numpy.vstack((fdx(*params), fdp(*params)))
            for argidx in range(len(deriv)):
                pmeps = copy.copy(params)
                ppeps = copy.copy(params)
                pmeps[argidx] = pmeps[argidx] - eps / 2
                ppeps[argidx] = ppeps[argidx] + eps / 2
                findiff = (f(*ppeps) - f(*pmeps)) / eps
                diff = numpy.abs(deriv[argidx] - findiff)
                errinfo = f'maximum difference, idx/npoints: ' \
                    f'{numpy.max(diff)}, {numpy.argmax(diff)}/{len(diff)}\n' \
                    f'parameters: {str(p)}'
                self.assertTrue(numpy.allclose(deriv[argidx], findiff,
                                               atol=1e3*eps),
                                f'{str(f)}, {argidx}, derivatives not close '
                                f'to numerical approximation ({errinfo})')


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

import os.path
import tempfile
import unittest

import numpy
import xrayutilities as xu

testfile = 'p08_00019.FIO'
datadir = os.path.join(os.path.dirname(__file__), 'data')
fullfilename = os.path.join(datadir, testfile)


@unittest.skipIf(not os.path.isfile(fullfilename),
                 "additional test data needed (http://xrayutilities.sf.io)")
class TestIO_FIO(unittest.TestCase):
    dshape = (601,)
    dmax = 2272108.0
    dmin = 66133.0
    dmaxcorr = 216583893672.0
    dmincorr = 0.0
    motmax = 8.000001
    motmin = 0.0
    tpos = 300
    dtpos = 1278952.0
    motorname = 'Motor_TT'
    countername = 'MythenIntegral'
    P08_normalizer = xu.IntensityNormalizer(
        "MCA",
        time='CountingTime',
        mon='MonitorEnergyWindow',
        absfun=lambda d: d['AttenuationFactor'])

    @classmethod
    def setUpClass(cls):
        scanname = os.path.splitext(testfile)[0]
        mcatmp = os.path.join(datadir, scanname, scanname+"_mythen_%i.raw")
        cls.fiofile = xu.io.SPECTRAFile(fullfilename, mcatmp=mcatmp)
        cls.sdata = cls.fiofile.data
        cls.motor = cls.sdata[cls.motorname]
        cls.inte = cls.sdata[cls.countername]
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, 'tmp.h5')
            cls.fiofile.Save2HDF5(fname, scanname)
            [cls.h5tt, _, cls.h5int], cls.h5data = xu.io.geth5_spectra_map(
                fname, [19], cls.motorname, 'ZS', cls.countername)

    def test_datashape(self):
        self.assertEqual(self.dshape, self.sdata[self.motorname].shape)
        self.assertEqual(self.dshape, self.sdata[self.countername].shape)

    def test_datavalues(self):
        self.assertAlmostEqual(self.motmax, self.motor.max(), places=6)
        self.assertAlmostEqual(self.motmin, self.motor.min(), places=6)
        self.assertAlmostEqual(self.dmax, self.inte.max(), places=6)
        self.assertAlmostEqual(self.dmin, self.inte.min(), places=6)
        self.assertAlmostEqual(self.dtpos, self.inte[self.tpos], places=6)

    def test_hdf5file(self):
        self.assertTrue(numpy.all(self.inte == self.h5int))
        self.assertTrue(numpy.all(self.motor == self.h5tt))

    def test_normalizer(self):
        mcac = self.P08_normalizer(self.h5data)
        self.assertAlmostEqual(self.dmaxcorr, mcac.max(), places=6)
        self.assertAlmostEqual(self.dmincorr, mcac.min(), places=6)


if __name__ == '__main__':
    unittest.main()



"""
example script to show the detector parameter determination for area detectors
from images recorded in the primary beam
"""

import os

import xrayutilities as xu

en = 10300.0  # eV
datadir = os.path.join("data", "wire_")  # data path for CCD files
# template for the CCD file names
filetmp = os.path.join(datadir, "wire_12_%05d.edf.gz")

# manually selected images
# select images which have the primary beam fully on the CCD
imagenrs = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
            20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33]

images = []
ang1 = []
ang2 = []

# read images and angular positions from the data file
# this might differ for data taken at different beamlines since
# they way how motor positions are stored is not always consistent
for imgnr in imagenrs:
    filename = filetmp % imgnr
    edf = xu.io.EDFFile(filename)
    images.append(edf.data)
    ang1.append(float(edf.header['ESRF_ID01_PSIC_NANO_NU']))
    ang2.append(float(edf.header['ESRF_ID01_PSIC_NANO_DEL']))


# call the fit for the detector parameters
# detector arm rotations and primary beam direction need to be given.
# in total 9 parameters are fitted, however the severl of them can
# be fixed. These are the detector tilt azimuth, the detector tilt angle, the
# detector rotation around the primary beam and the outer angle offset
# The detector pixel size or the detector distance should be kept unfixed to
# be optimized by the fit.
param, eps = xu.analysis.sample_align.area_detector_calib(
    ang1, ang2, images, ['z+', 'y-'], 'x+',
    start=(None, None, 1.0, 45, 0, -0.7, 0),
    fix=(False, False, True, False, False, False, False),
    wl=xu.en2lam(en))



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu


class TestAlloyContentCalc(unittest.TestCase):
    matA = xu.materials.InAs
    matB = xu.materials.InP
    substrate = xu.materials.InAs
    x = numpy.random.rand()
    alloy = xu.materials.CubicAlloy(matA, matB, x)
    hxrd001 = xu.HXRD([1, 1, 0], [0, 0, 1])
    [qxa, qza] = alloy.RelaxationTriangle([2, 2, 4], substrate, hxrd001)
    [qxs, qzs] = alloy.RelaxationTriangle([0, 0, 4], substrate, hxrd001)

    def test_ContentAsym(self):
        content, [ainp, aperp, abulk, eps_inp, eps_perp] = \
            self.alloy.ContentBasym(self.qxa[0], self.qza[0],
                                    [2, 2, 4], [0, 0, 1])
        self.assertAlmostEqual(content, self.x, places=6)
        content, [ainp, aperp, abulk, eps_inp, eps_perp] = \
            self.alloy.ContentBasym(self.qxa[1], self.qza[1],
                                    [2, 2, 4], [0, 0, 1])
        self.assertAlmostEqual(content, self.x, places=6)

    def test_ContentSym(self):
        content = self.alloy.ContentBsym(
            self.qzs[0], [0, 0, 4], [1, 1, 0], self.matA.lattice.a, 1.0)
        self.assertAlmostEqual(content, self.x, places=6)
        content = self.alloy.ContentBsym(
            self.qzs[1], [0, 0, 4], [1, 1, 0], self.matA.lattice.a, 0.0)
        self.assertAlmostEqual(content, self.x, places=6)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2013, 2015 Dominik Kriegner <dominik.kriegner@gmail.com>

# module for handling files stored in the CBF data format

import os.path
import re

import numpy

from .. import config, cxrayutilities, utilities
from .filedir import FileDirectory
from .helper import xu_h5open, xu_open

cbf_name_start_num = re.compile(r"^\d")


class CBFFile(object):

    def __init__(self, fname, nxkey="X-Binary-Size-Fastest-Dimension",
                 nykey="X-Binary-Size-Second-Dimension",
                 dtkey="DataType", path=None):
        """
        CBF detector image parser

        Parameters
        ----------
        fname :     str
            name of the CBF file of type .cbf or .cbf.gz
        nxkey :     str, optional
            name of the header key that holds the number of points in
            x-direction
        nykey :     str, optional
            name of the header key that holds the number of points in
            y-direction
        dtkey :     str, optional
            name of the header key that holds the datatype for the binary data
        path :      str, optional
            path to the CBF file
        """

        self.filename = fname
        if path:
            self.full_filename = os.path.join(path, fname)
        else:
            self.full_filename = self.filename

        # evaluate keyword arguments
        self.nxkey = nxkey
        self.nykey = nykey
        self.dtkey = dtkey

        # create attributes for holding data
        self.data = None
        self.ReadData()

    def ReadData(self):
        """
        Read the CCD data into the .data object
        this function is called by the initialization
        """
        with xu_open(self.full_filename, 'rb') as fid:
            tmp = numpy.fromfile(file=fid, dtype="u1").tobytes()
            tmp2 = tmp.decode('ascii', 'ignore')
            # read header information
            pos = tmp2.index(self.nxkey + ':') + len(self.nxkey + ':')
            self.xdim = int(tmp2[pos:pos + 6].strip())
            pos = tmp2.index(self.nykey + ':') + len(self.nykey + ':')
            self.ydim = int(tmp2[pos:pos + 6].strip())

            self.data = cxrayutilities.cbfread(tmp, self.xdim, self.ydim)
            self.data.shape = (self.ydim, self.xdim)

    def Save2HDF5(self, h5f, group="/", comp=True):
        """
        Saves the data stored in the EDF file in a HDF5 file as a HDF5 array.
        By default the data is stored in the root group of the HDF5 file - this
        can be changed by passing the name of a target group or a path to the
        target group via the "group" keyword argument.

        Parameters
        ----------
        h5f :   file-handle or str
            a HDF5 file object or name
        group : str, optional
            group where to store the data (default to the root of the file)
        comp :  bool, optional
            activate compression - true by default
        """
        with xu_h5open(h5f, 'a') as h5:
            if isinstance(group, str):
                g = h5.get(group)
            else:
                g = group

            # create the array name
            name = os.path.split(self.filename)[-1]
            name = os.path.splitext(name)[0]
            # perform a second time for case of .cbf.gz files
            name = os.path.splitext(name)[0]
            name = utilities.makeNaturalName(name)
            if cbf_name_start_num.match(name):
                name = "ccd_" + name
            if config.VERBOSITY >= config.INFO_ALL:
                print("xu.io.CBFFile: HDF5 group name: %s" % name)

            # create the array description
            desc = "CBF CCD data from file %s " % (self.filename)

            # create the dataset for the array
            kwds = {'fletcher32': True}
            if comp:
                kwds['compression'] = 'gzip'

            try:
                ca = g.create_dataset(name, data=self.data, **kwds)
            except ValueError:
                del g[name]
                ca = g.create_dataset(name, data=self.data, **kwds)

            ca.attrs['TITLE'] = desc


class CBFDirectory(FileDirectory):

    """
    Parses a directory for CBF files, which can be stored to a HDF5 file for
    further usage
    """

    def __init__(self, datapath, ext="cbf", **keyargs):
        """
        Parameters
        ----------
        datapath :  str
            directory of the CBF files
        ext :       str, optional
            extension of the ccd files in the datapath (default: "cbf")
        keyargs :   dict, optional
            further keyword arguments are passed to CBFFile
        """
        super().__init__(datapath, ext, CBFFile, **keyargs)
